from utils import move_to
from .algorithm import Algorithm
import torch


class EncoderDecoderAlgorithm(Algorithm):
    def __init__(self, config, datasets):
        super().__init__(config, datasets)
        self.generation_config = config.generation_config
        self.generate_during_training = config.generate_during_training

    def process_batch(self, batch):
        """
        A helper function for update() and evaluate() that process the batch
        Args:
            - batch: a batch of data yielded by the DataLoader
        Output:
            - results: a dictionary of results
                - y_pred: the prediction of the model
                - y_true: the ground truth
                - metadata: the metadata of the batch
                - objective: a dictionary with the loss name and loss value
        """
        X, y_true, metadata = batch
        
        # task-specific preprocessing
        X, y_true, metadata = getattr(self, f"_{metadata['task_metadata']['type']}_preprocessing")(X, y_true, metadata)

        # prepare inputs for generation if necessary
        if self.requires_metric_calculation():
            self.model.tokenizer.padding_side="left"
            X_generate = self.model.transform_inputs(X).input_ids
            X_generate = move_to(X_generate, self.device)
            self.model.tokenizer.padding_side="right"

        X = self.model.transform_inputs(X)
        lm_labels = self.model.transform_outputs(y_true)

        X['lm_labels'] = lm_labels
        X = move_to(X, self.device)

        loss = self.model(**X)

        # generate predictions and convert to labels if necessary
        if self.requires_metric_calculation():
            # generate predictions
            outputs = self.model.generate(X_generate, **self.generation_config)
            
            # task-specific postprocessing
            y_pred, y_true = getattr(self, f"_{metadata['task_metadata']['type']}_postprocessing")(outputs, y_true, metadata)

        else:
            y_pred = []
            y_true = []

        results = {
            'y_pred': y_pred,
            'y_true': y_true,
            'metadata': metadata,
            "objective": {
                "loss_name": "LM_cross_entropy",
                "loss_value": loss.item()}
        }

        return results, loss

    def requires_metric_calculation(self):
        # determines whether the model needs to generate predictions
        #       else only calculates loss
        if self.is_training and not self.generate_during_training:
            return False
        return True

    def _classification_preprocessing(self, X, y_true, metadata):
        return X, y_true, metadata

    def _classification_postprocessing(self, outputs, y_true, metadata):
        outputs = self.model.batch_decode(outputs)
        y_true = self.convert_strings_to_labels(metadata['labels'], y_true)
        assert(all(y_true != -1)),str(y_true)
        y_pred = self.convert_strings_to_labels(metadata['labels'], outputs)
        
        return y_pred, y_true

    def _multioutput_classification_preprocessing(self, X, y_true, metadata):
        return X, y_true, metadata

    def _multioutput_classification_postprocessing(self, outputs, y_true, metadata):
        outputs = self.model.batch_decode(outputs)
        y_true = self.convert_strings_to_labels(metadata['labels'], y_true)
        assert(all(y_true != -1)),str(y_true)
        y_pred = self.convert_strings_to_labels(metadata['labels'], outputs)
        return y_pred, y_true

    def _multilabel_classification_preprocessing(self, X, y_true, metadata):
        # format y_true into a string of labels
        y_true = [" and ".join([metadata['task_metadata']['class_to_natural_language_map'][c] for c in sample]) for sample in y_true]
        return X, y_true, metadata

    def _multilabel_classification_postprocessing(self, outputs, y_true, metadata):
        outputs = self.model.batch_decode(outputs)
        # convert model outputs to mutlilabel format
        y_pred = []
        for output in outputs:
            pred = [0 for _ in range(len(metadata['labels']))]
            # search for class names in output
            for i, natural_language_class in enumerate(metadata['task_metadata']['class_to_natural_language_map'].values()):
                if natural_language_class in output:
                    prediction = list(metadata['task_metadata']['class_to_natural_language_map'].keys())[i]
                    pred[i] = 1
            if sum(pred) == 0:
                pred[metadata['labels'].index(metadata['task_metadata']['default_prediction'])] = 1
            y_pred.append(pred)

        # convert labels to multilabel format
        transformed_y_true = []
        for y in y_true:
            true = [0 for _ in range(len(metadata['labels']))]
            natural_language_labels = y.split(" and ")
            label_indices = [list(metadata['task_metadata']['class_to_natural_language_map'].values()).index(l) for l in natural_language_labels]
            for i in label_indices:
                true[i] = 1
            transformed_y_true.append(true)

        return y_pred, transformed_y_true

    def _span_extraction_preprocessing(self, X, y_true, metadata):
        if isinstance(y_true[0], list):
            y_true = [[y_['text'] for y_ in y] for y in y_true]
        else:
            y_true = [y['text'] for y in y_true]
        return X, y_true, metadata

    def _span_extraction_postprocessing(self, outputs, y_true, metadata):
        y_pred = self.model.batch_decode(outputs)
        return y_pred, y_true

    def _multiple_choice_preprocessing(self, X, y_true, metadata):
        return X, y_true, metadata

    def _multiple_choice_postprocessing(self, outputs, y_true, metadata):
        outputs = self.model.batch_decode(outputs)
        num_choices = metadata['task_metadata']['num_choices']
        metadata['labels'] = [str(i) for i in range(num_choices)]
        y_true = self.convert_strings_to_labels(metadata['labels'], y_true)
        assert(all(y_true != -1)),str(y_true)
        y_pred = self.convert_strings_to_labels(metadata['labels'], outputs)
        return y_pred, y_true

    def _response_generation_preprocessing(self, X, y_true, metadata):
        return X, y_true, metadata

    def _response_generation_postprocessing(self, outputs, y_true, metadata):
        y_pred = self.model.batch_decode(outputs)
        return y_pred, y_true


from torch.utils.data import DataLoader
import numpy as np
import random

def get_train_loader(dataset, batch_size, config, **loader_kwargs):
    """
    Constructs and return the data loader for training
    Args:
        - dataset (TLiDBDataset): The dataset to load the data from
        - batch_size (int): The batch size for the data loader
        - **loader_kwargs (dict): The keyword arguments for the data loader
    Returns:
        - data_loader (DataLoader): The data loader for training
    """
    if dataset.task_metadata['type'] == "multiple_choice" and config.model_type == "Encoder":
        # Encoder-only models split multiple choice into num_choices samples
        #   so we need to downscale the batch_size accordingly
        batch_size = batch_size // dataset.task_metadata['num_choices']
    if loader_kwargs['num_workers'] > 1:
        loader_kwargs['pin_memory'] = True
    return DataLoader(dataset, batch_size=batch_size, shuffle=True, **loader_kwargs)

def get_eval_loader(dataset, batch_size, config, **loader_kwargs):
    """
    Constructs and return the data loader for evaluation
    Args:
        - dataset (TLiDBDataset): The dataset to load the data from
        - batch_size (int): The batch size for the data loader
        - **loader_kwargs (dict): The keyword arguments for the data loader
    Returns:
        - data_loader (DataLoader): The data loader for evaluation
    """
    if dataset.task_metadata['type'] == "multiple_choice" and config.model_type == "Encoder":
        # Encoder-only models split multiple choice into num_choices samples
        #   so we need to downscale the batch_size accordingly
        batch_size = batch_size // dataset.task_metadata['num_choices']
    if loader_kwargs['num_workers'] > 1:
        loader_kwargs['pin_memory'] = True
    return DataLoader(dataset, batch_size=batch_size, shuffle=False, **loader_kwargs)

def get_loader(split):
    if split == 'train':
        return get_train_loader
    else:
        return get_eval_loader

class TLiDB_DataLoader:
    """
    Data loader that combines and samples from multiple single-task data loaders

    Args:
        - dataloader_dict (dict): A dictionary containing the following keys:
            - datasets (list): A list of TLiDB_Dataset objects
            - loaders (list): A list of DataLoader objects
            - losses (list): A list of strings corresponding to the name of the loss for each dataset
    Returns:
        - batch (tuple): A tuple containing the following elements:
            - X (str): The input data
            - y (str): The target data
            - metadata (dict): A dictionary containing the following keys:
                - task (str): The name of the task
                - dataset_name (str): The name of the dataset
                - loss (str): The name of the loss
    """
    def __init__(self, dataloader_dict):
        self.dataset_names = [d.dataset_name for d in dataloader_dict['datasets']]
        self.task_names = [d.task for d in dataloader_dict['datasets']]
        self.dataloaders = [iter(d) for d in dataloader_dict['loaders']]
        self.lengths = [len(d) for d in self.dataloaders]
        self.remaining_batches = [len(d) for d in self.dataloaders]
        self.task_metadatas = [d.task_metadata for d in dataloader_dict['datasets']]
        self.task_weights = [1/l**(1/2) for l in self.lengths]

    def _reset(self):
        pass

    def __iter__(self):
        self._reset()
        return self
    
    def __len__(self):
        return sum(self.lengths)

    def __next__(self):
        if sum(self.remaining_batches) > 0:
            # sample probabilistically from each data loader
            selected_idx = random.choices(range(len(self.dataloaders)), weights=self.remaining_batches)[0]
            selected_loader = self.dataloaders[selected_idx]
            self.remaining_batches[selected_idx] -= 1
            batch = next(selected_loader)
            X, y, metadata = batch
            metadata['task'] = self.task_names[selected_idx]
            metadata['dataset_name'] = self.dataset_names[selected_idx]
            metadata['task_metadata'] = self.task_metadatas[selected_idx]
            metadata['task_weight'] = self.task_weights[selected_idx]
            return X,y,metadata
        else:
            raise StopIteration


import os
import json
from urllib.request import urlopen
from io import BytesIO
from zipfile import ZipFile
import numpy as np
from torch.utils.data import Dataset


def download_and_unzip(url, extract_to='.'):
    print(f"Waiting for response from {url}")
    http_response = urlopen(url)
    print(f"Downloading data from {url}")
    zipfile = ZipFile(BytesIO(http_response.read()))
    zipfile.extractall(path=extract_to)

def load_dataset_local(name, dataset_folder):
    ds = {}
    for root, dirs, files in os.walk(f"{dataset_folder}/TLiDB_{name}"):
        for file in files:
            if file.endswith(".json") and file!="sample_format.json":
                ds[file[:-5]] = json.load(open(f"{root}/{file}"))
    
    if len(ds.keys()) == 1:
        ds = ds[list(ds.keys())[0]]
    return ds

def load_dataset(name, dataset_folder, url):
    # download and unzip dataset if needed
    if f"TLiDB_{name}" not in os.listdir(dataset_folder):
        assert(url is not None), "Must provide a url to download from"
        download_and_unzip(url, dataset_folder)
        print(f"Extracted files to {os.path.join(dataset_folder,name)}")

    ds = load_dataset_local(name, dataset_folder)

    return ds

def load_split_ids(name, dataset_folder, split, few_shot_percent=None):
    if f"TLiDB_{name}" not in os.listdir(dataset_folder):
        raise ValueError("Dataset not found")
    if few_shot_percent and split!="test":
        ids_file = f"{dataset_folder}/TLiDB_{name}/TTiDB_{few_shot_percent}_percent_few_shot_{split}_ids.txt"
    else:
        ids_file = f"{dataset_folder}/TLiDB_{name}/TTiDB_{split}_ids.txt"
    with open(ids_file) as f:
        ids = f.read().splitlines()
    return ids

class TLiDB_Dataset(Dataset):
    """
    Abstract dataset class for all TLiDB datasets
    """
    def __init__(self, dataset_name, task, model_type, max_dialogue_length, dataset_folder):
        super().__init__()
        self.dataset = load_dataset(dataset_name, dataset_folder, self.url)
        self._task = task
        task_metadata = self.dataset['metadata']['task_metadata']
        self.task_labels = []
        self._max_dialogue_length = max_dialogue_length
        self._model_type = model_type
        if task in task_metadata and 'labels' in task_metadata[task]:
            self.task_labels = task_metadata[task]['labels']

        if task == "response_generation":
            self.metrics = ['token_f1']
            self.metric_kwargs = {}
            self._collate = self._collate_response_generation
        else:
            self.metrics = task_metadata[task]['metrics']
            self.metric_kwargs = task_metadata[task].get("metric_kwargs", dict())

            if model_type == "Encoder":
                self._collate = self._collate_encoder
            elif model_type == "Decoder":
                self._collate = self._collate_decoder
            elif model_type == "EncoderDecoder":
                self._collate = self._collate_encoderdecoder
            else:
                raise ValueError(f"{model_type} is not a valid algorithm type")

    @property
    def dataset_name(self):
        return self._dataset_name
    
    @property
    def tasks(self):
        return self._tasks

    @property
    def task(self):
        return self._task

    @property
    def task_metadata(self):
        return self._task_metadata

    @property
    def url(self):
        return self._url

    @property
    def max_dialogue_length(self):
        return self._max_dialogue_length

    @property
    def model_type(self):
        return self._model_type

    @property
    def collate(self):
        """
        Returns collate function to be used with dataloader
        By default returns None -> uses default torch collate function
        """
        return getattr(self, "_collate", None)

    @property
    def y_array(self):
        """
        Targets for the model to predict, can be labels for classification or string for generation tasks
        """
        return self._y_array

    @property
    def y_size(self):
        """
        Number of elements in the target
        For standard classification and text generation, y_size = 1
        For multi-class or multi-task prediction tasks, y_size > 1
        """
        return self._y_size

    @property
    def num_classes(self):
        """
        Returns number of classes in the dataset
        """
        return getattr(self, "_num_classes", None)

    @property
    def metadata_fields(self):
        """
        Returns the fields that are stored in the metadata
        Metadata should always contain the domains
        If it is a classification task, then metadata should also contain the classes
        """
        return self._metadata_fields

    @property
    def metadata_array(self):
        """
        Returns the metadata array
        """
        return self._metadata_array

    def get_metadata_field(self, field):
        return self.metadata_array[self.metadata_fields.index(field)]

    def __getitem__(self, idx):
        """
        Returns a single sample from the dataset
        """
        x = self.get_input(idx)
        y = self.get_target(idx)
        m = self.get_metadata(idx)
        return x, y, m

    def get_input(self, idx):
        return self._input_array[idx]

    def get_target(self, idx):
        return self.y_array[idx]

    def get_metadata(self, idx):
        return {}

    def _collate(self, batch):
        return NotImplementedError

    def _collate_encoder(self, batch):
        return NotImplementedError
    
    def _collate_decoder(self, batch):
        return NotImplementedError

    def _collate_encoderdecoder(self, batch):
        return NotImplementedError

    def __len__(self):
        """
        Returns the length of the dataset
        """
        return len(self.y_array)

    def _truncate_dialogue(self, input):
        """
        Truncates the dialogue to the max dialogue length
        """
        if self.max_dialogue_length:
            dialogue = self._convert_dialogue_to_string(input)
            while len(dialogue.split()) > self.max_dialogue_length:
                input = input[1:]
                dialogue = self._convert_dialogue_to_string(input)

        return input

    def _convert_dialogue_to_string(self, input):
        dialogue = ""
        for (speaker, utt) in input:
            if speaker:
                dialogue += f"{speaker}: "
            dialogue += f"{utt} "
        return dialogue[:-1]

    def _join_strings(self, *args):
        return " ".join(args)

    def _load_response_generation_task(self, task, split_ids):
        for datum in self.dataset['data']:
            if datum['dialogue_id'] in split_ids:
                dialogue_str = ""
                for turn in datum['dialogue']:
                    dialogue_str += f"{turn['speakers'][0]}: "
                    self._input_array.append(dialogue_str)
                    self._y_array.append(turn['utterance'])
                    dialogue_str += f"{turn['utterance']} "

    def _collate_response_generation(self, batch):
        X, y, metadata = [], [], {}
        for item in batch:
            X.append(item[0])
            y.append(item[1])
            for k, v in item[2].items():
                if k not in metadata:
                    metadata.append(k)
                metadata[k].append(v)
        return X, y, metadata

    def random_subsample(self, frac=1.0):
        """
        Subsamples the dataset
        Args:
            - frac (float): Fraction of the dataset to keep
        """
        if frac < 1.0:
            num_to_retain = int(self.y_size * frac)
            if num_to_retain == 0:
                return
            idxs_to_retain = np.sort(np.random.permutation(len(self))[:num_to_retain]).tolist()
            subsampled_input_array, subsampled_y_array, subsampled_metadata_array = [], [], []
            for idx in idxs_to_retain:
                input_item, y_item, metadata_item = self.__getitem__(idx)
                subsampled_input_array.append(input_item)
                subsampled_y_array.append(y_item)
                subsampled_metadata_array.append(metadata_item)
            self._input_array = subsampled_input_array
            self._y_array = subsampled_y_array
            metadata_iterated = list(metadata_item.keys())
            metadata_not_iterated = [metadata_field for metadata_field in self.metadata_fields if metadata_field not in metadata_iterated]
            subsampled_metadata_array = [subsampled_metadata_array]
            for metadata_field in metadata_not_iterated:
                subsampled_metadata_array.append(self.get_metadata_field(metadata_field))
            self._metadata_array = subsampled_metadata_array
            self._metadata_fields = metadata_iterated+metadata_not_iterated
            self._y_size = num_to_retain



import torch
import torch.nn as nn

from examples.utils import move_to, detach_and_clone
from examples.optimizers import initialize_optimizer
from examples.models.initializer import initialize_model

class Algorithm(nn.Module):
    def __init__(self, config, datasets):
        super().__init__()
        self.model = initialize_model(config, datasets)
        self.model.to(config.device)
        self.device = config.device
        self.out_device = 'cpu'
        self.optimizer = initialize_optimizer(config, self.model)
        self.max_grad_norm = config.max_grad_norm
        self.gradient_accumulation_steps = max(config.effective_batch_size//config.gpu_batch_size,1)
        self.imbalanced_task_weighting = config.imbalanced_task_weighting
        if not (config.device == 'cpu') and config.fp16:
            self.fp16 = config.fp16
            self.scaler = torch.cuda.amp.GradScaler()
        else:
            self.fp16 = False
    
    def process_batch(self, batch):
        raise NotImplementedError

    @property
    def requires_metric_calculation(self):
        """Whether to calculate metrics"""
        return NotImplementedError
    
    def state_dict(self):
        return self.model.state_dict()

    def load_state_dict(self, state_dict):
        """
        Load the state dict of the model
        """
        self.model.load_state_dict(state_dict)

    def update(self, batch, step):
        """
        Process the batch, and update the model
        Args:
            - batch: a batch of data yielded by data loaders
        Output:
            - results (dict): information about the batch, such as:
                - y_pred: the predicted labels
                - y_true: the true labels
                - metadata: the metadata of the batch
                - loss: the loss of the batch
                - metrics: the metrics of the batch
        """
        assert self.is_training, "Cannot update() when not in training mode"
        
        if self.fp16:
            with torch.cuda.amp.autocast():
                results, objective = self.process_batch(batch)
                if self.imbalanced_task_weighting:
                    task_weight = torch.tensor(batch[2]['task_weight']).to(self.device)
                    objective = task_weight * objective
                objective = objective / self.gradient_accumulation_steps
            self.scaler.scale(objective).backward()
            if ((step+1)%self.gradient_accumulation_steps) == 0:
                self.scaler.unscale_(self.optimizer)
                nn.utils.clip_grad_norm_(self.model.parameters(), self.max_grad_norm)
                self.scaler.step(self.optimizer)
                self.scaler.update()
                self.model.zero_grad(set_to_none=True)
        else:
            results, objective = self.process_batch(batch)
            if self.imbalanced_task_weighting:
                task_weight = torch.tensor(batch[2]['task_weight']).to(self.device)
                objective = task_weight * objective
            objective = objective / self.gradient_accumulation_steps
            objective.backward()
            if ((step+1)%self.gradient_accumulation_steps) == 0:
                nn.utils.clip_grad_norm_(self.model.parameters(), self.max_grad_norm)
                self.optimizer.step()
                self.model.zero_grad(set_to_none=True)

        return self.sanitize_dict(results)

    def evaluate(self, batch):
        """
        Process the batch, and evaluate the model
        Args:
            - batch: a batch of data yielded by data loaders
        Output:
            - results (dict): information about the batch, such as:
                - y_pred: the predicted labels
                - y_true: the true labels
                - metadata: the metadata of the batch
                - loss: the loss of the batch
                - metrics: the metrics of the batch
        """
        assert not self.is_training, "Cannot evaluate() when in training mode"
        with torch.no_grad():
            results, _ = self.process_batch(batch)
        return self.sanitize_dict(results)

    def train(self, mode=True):
        """
        Set the model to training mode
        """
        self.is_training = mode
        super().train(mode)

    def eval(self):
        """
        Set the model to evaluation mode
        """
        self.train(False)

    def sanitize_dict(self, in_dict, to_out_device=True):
        """
        Helper function that sanitizes dictionaries by:
            - moving to the specified output device
            - removing any gradient information
            - detaching and cloning the tensors
        Args:
            - in_dict (dictionary)
        Output:
            - out_dict (dictionary): sanitized version of in_dict
        """
        out_dict = detach_and_clone(in_dict)
        if to_out_device:
            out_dict = move_to(out_dict, self.out_device)
        return out_dict

    def convert_strings_to_labels(self, labels, strings):
        return torch.tensor([labels.index(s) if s in labels else -1 for s in strings])


import argparse
import torch
import configs

incompatible_with_fp16 = ["t5-base"]
incompatible_with_generation = ["bert"]

def parse_args():
    parser = argparse.ArgumentParser()
    
    # debugging args
    parser.add_argument("--frac", type=float, default=1.0,
        help="Convenience parameter that scales down dataset size to specified fraction, for debugging purposes")
    parser.add_argument("--debug", action="store_true", help="Setup training for debugging. For example, no logging")
    parser.add_argument("--generate_during_training", action="store_true")

    # configs for experimentation ease
    parser.add_argument("--model_config",type=str, default=None)
    parser.add_argument("--num_workers", type=int, default=4)

    # general args
    parser.add_argument("--cpu_only", action="store_true")
    parser.add_argument("--seed", type=int, default=-1)
    parser.add_argument("--log_and_model_dir", type=str, default="./logs_and_models")
    parser.add_argument("--saved_model_dir", type=str, default=None, help="To load a saved model for fine-tuning or evaluation")
    parser.add_argument("--data_dir", type=str, default="../TLiDB/data",
        help="The directory where data is stored, by default we place it inside the TLiDB package in a data folder")

    # model args
    parser.add_argument("--model", type=str)
        
    # training args
    parser.add_argument("--do_train", action="store_true")
    parser.add_argument("--do_finetune", action="store_true")
    parser.add_argument("-e", "--num_epochs", type=int, default=10)
    parser.add_argument("--effective_batch_size", type=int, default=60)
    parser.add_argument("--gpu_batch_size", type=int, default=20)
    parser.add_argument("-lr", "--learning_rate", type=float, default=3e-5)
    parser.add_argument("--fp16", action="store_true")
    parser.add_argument("--max_grad_norm", type=float, default=1.0)
    parser.add_argument("--save_best", type=bool, default=True)
    parser.add_argument("--save_last", type=bool, default=False)
    parser.add_argument("--imbalanced_task_weighting", type=bool, default=True)

    # evaluation args
    parser.add_argument("--do_eval", action="store_true")
    parser.add_argument("--eval_best", action="store_true")
    parser.add_argument("--eval_last", action="store_true")
    
    # task args
    parser.add_argument("--source_tasks", type=str,nargs='+',required=True)
    parser.add_argument("--source_datasets", type=str,nargs='+',required=True)
    parser.add_argument("--target_tasks", type=str,nargs='+')
    parser.add_argument("--target_datasets", type=str,nargs='+')

    # TTiDB args
    parser.add_argument("--cotraining", action="store_true")
    parser.add_argument("--few_shot_percent", type=float, default=None)
    
    # algorithm args
    parser.add_argument("--optimizer", type=str, default="Adam")
    parser.add_argument("--weight_decay", type=float, default=0.0)

    # misc. args
    parser.add_argument("--progress_bar", type=bool, default=True)
    parser.add_argument("--save_pred",action="store_true",help="Save predictions for each task")
    parser.add_argument("--resume", action="store_true", help="Resume training from checkpoint")

    args = parser.parse_args()

    if args.debug:
        # send debugging logs and models to different directory
        args.log_and_model_dir = "./debug_logs_and_models"

    if args.model_config is not None:
        for attr, value in configs.__dict__[f"{args.model_config}_config"].items():
            setattr(args, attr, value)

    if "bert" in args.model:
        setattr(args, "output_type", "categorical")
        setattr(args, "model_type", "Encoder")
    elif "gpt" in args.model:
        setattr(args, "output_type", "token")
        setattr(args, "model_type", "Decoder")
        setattr(args, "generation_config", configs.GPT2_generation_config)
    elif "t5" in args.model:
        setattr(args, "output_type", "token")
        setattr(args, "model_type", "EncoderDecoder")
        setattr(args, "generation_config", configs.t5_generation_config)
    else:
        raise ValueError(f"Model {args.model} not supported")

    if not args.cpu_only:
        setattr(args, "device", "cuda" if torch.cuda.is_available() else "cpu")
    else:
        setattr(args, "device", "cpu")

    if args.num_workers > 1:
        import os
        os.environ['TOKENIZERS_PARALLELISM'] = "true"

    assert(not(args.model in incompatible_with_fp16 and args.fp16)), f"Cannot use fp16 with model {args.model}"
    assert(not(args.model in incompatible_with_generation and\
            any(['generation' in t for t in args.source_tasks+args.target_tasks]))), f"Cannot perform generation tasks with model {args.model}"

    return args



from algorithms.initializer import initialize_algorithm
from train import train,evaluate
import os
from utils import Logger, load_datasets_split, load_algorithm, log_config, \
        set_seed, log_dataset_info, get_savepath_dir, append_to_save_path_dir
import argparser

def main(config):

    # if cotraining, then train on both source+target tasks, and dev is target only
    if config.cotraining:
        config.train_datasets = config.source_datasets+config.target_datasets
        config.train_tasks = config.source_tasks+config.target_tasks
        config.dev_datasets = config.target_datasets
        config.dev_tasks = config.target_tasks
    # if training only on source tasks, then train/dev are the same
    else:
        config.train_datasets = config.source_datasets
        config.train_tasks = config.source_tasks
        config.dev_datasets = config.source_datasets
        config.dev_tasks = config.source_tasks

    if config.target_datasets and config.target_tasks:
        # always finetune and evaluate on target tasks
        config.finetune_datasets = config.target_datasets
        config.finetune_tasks = config.target_tasks
        config.eval_datasets = config.target_datasets
        config.eval_tasks = config.target_tasks

    # create save path based only on train data
    config.save_path_dir = get_savepath_dir(config.train_datasets, config.train_tasks, config.seed, config.log_and_model_dir, config.model, config.few_shot_percent, config.cotraining)

    # Initialize logs
    if os.path.exists(config.save_path_dir) and \
        (config.resume or ((config.do_finetune or config.do_eval) and (not config.do_train))):
        # if explicitly resuming, or running eval only then append to logger
        resume=True
        mode='a'
    else:
        resume=False
        mode='w'

    if not os.path.exists(config.save_path_dir):
        os.makedirs(config.save_path_dir)

    if config.debug:
        logger = Logger(mode='w')
    else:
        logger = Logger(os.path.join(config.save_path_dir, 'log.txt'), mode)

    set_seed(config.seed)
    if config.do_train:
        datasets = {}

        # load datasets for training
        datasets['train'] = load_datasets_split("train",config.train_tasks, config.train_datasets, config)
        datasets['dev'] = load_datasets_split("dev",config.dev_tasks, config.dev_datasets, config)

        # log configuration and dataset info
        logger.write("TRAINING\n")
        log_config(config,logger)
        log_dataset_info(datasets, logger)

        # initialize algorithm
        algorithm = initialize_algorithm(config, datasets)

        # try to resume training from a saved model
        resume_success = False
        if resume:
            if os.path.exists(os.path.join(config.save_path_dir, 'last_model.pt')):
                prev_epoch, best_val_metric = load_algorithm(algorithm, os.path.join(config.save_path_dir, 'last_model.pt'),logger)
                epoch_offset = prev_epoch + 1
                logger.write(f"Resuming training from epoch {prev_epoch} with best validation metric {best_val_metric}\n")
                resume_success = True
            else:
                logger.write("No previous model found, starting from scratch\n")

        # if not resuming, or if resuming but no previous model found, then train from scratch
        if not resume_success:
            epoch_offset=0
            best_val_metric = None

        train(algorithm, datasets, config, logger, epoch_offset, best_val_metric)

    if config.do_finetune:
        assert(config.target_datasets and config.target_tasks),"Must specify target datasets and tasks to finetune"
        datasets = {}
        # get the pre-trained model path
        if config.do_train or (config.train_datasets and config.train_tasks):
            # Do nothing, this means we already have a save_dir_path and a model saved there
            pass
        elif config.saved_model_dir:
            # if user explcitily specified a pretrained model to finetune, then use that
            config.save_path_dir = config.saved_model_dir
        else:
            raise ValueError("To run fine-tuning, use:\n--saved_model_dir to specify the pre-trained model OR\
                \n--train_datasets and --train_tasks to specify the pretraining datasets and tasks")
        
        # if fine tuning, set fine-tune train, and fine-tune dev to the same tasks
        config.finetune_train_tasks = config.finetune_tasks
        config.finetune_train_datasets = config.finetune_datasets
        config.finetune_dev_tasks = config.finetune_tasks
        config.finetune_dev_datasets = config.finetune_datasets

        # load datasets for finetuning
        datasets['train'] = load_datasets_split("train",config.finetune_train_tasks,config.finetune_train_datasets,config)
        datasets['dev'] = load_datasets_split("dev", config.finetune_dev_tasks, config.finetune_dev_datasets, config)

       # initialize algorithm
        algorithm = initialize_algorithm(config, datasets)

        # always load best pretrained model
        model_path = os.path.join(config.save_path_dir, 'best_model.pt')
        is_best = True
        load_algorithm(algorithm, model_path, logger)
        epoch_offset = 0
        best_val_metric = None

        # update save path with fine-tuning details
        config.save_path_dir = append_to_save_path_dir(config.save_path_dir, config.finetune_datasets, config.finetune_tasks, config.few_shot_percent, config.seed)
        
        # note the fine-tuning in the pretrained model log
        logger.write(f"FINETUNING at {config.save_path_dir}\n")
        
        # create new logger for fine-tuning
        if not os.path.exists(config.save_path_dir):
            os.makedirs(config.save_path_dir)
        if config.debug:
            finetune_logger = Logger(mode='w')
        else:
            finetune_logger = Logger(os.path.join(config.save_path_dir, 'log.txt'), mode="w")

        # log configuration and dataset info
        finetune_logger.write("FINETUNING\n")
        finetune_logger.write(f"Loaded pretrained model from {model_path}\n")
        log_config(config,finetune_logger)
        log_dataset_info(datasets, finetune_logger)

        train(algorithm, datasets, config, finetune_logger, epoch_offset, best_val_metric)
        finetune_logger.close()

    if config.do_eval:
        assert(config.target_datasets and config.target_tasks),"Must specify target datasets and tasks to finetune"
        datasets = {}
        # If coming from training/fine-tuning, 
        #   this means we already have a save_dir_path from training/fine-tuning and a model saved there
        if config.do_finetune or config.do_train:
            pass
        elif config.saved_model_dir:
            # if user explcitily specified a model to evaluate, then use that
            config.save_path_dir = config.saved_model_dir
        elif (config.finetune_datasets and config.finetune_tasks) and (config.train_datasets and config.train_tasks):
            # Given all the datasets and tasks, we can infer the path to the fine-tuned model
            config.save_path_dir = append_to_save_path_dir(config.save_path_dir, config.finetune_datasets, config.finetune_tasks, config.few_shot_percent, config.seed)
        else:
            raise ValueError("To run evaluation, use:\n--saved_model_dir to specify the model to evaluate OR\
                \n--finetune_datasets and --finetune_tasks and --train_datasets and --train_tasks to infer the path to the model")

        # ensure user has specified a model to evaluate
        assert(not(config.eval_last and config.eval_best)), "cannot evaluate both last and best models"
        assert(config.eval_last or config.eval_best), "must evaluate at least one model"
        
        # create logger for evaluation
        if config.debug:
            eval_logger = Logger(mode='w')
        else:
            eval_logger = Logger(os.path.join(config.save_path_dir, 'log.txt'), mode="a")
        eval_logger.write("EVALUATING\n")
        
        # load datasets for evaluation
        datasets['test'] = load_datasets_split("test",config.eval_tasks, config.eval_datasets, config)

        # log configuration and dataset info
        log_config(config,eval_logger)
        log_dataset_info(datasets, eval_logger)

        # initialize algorithm
        algorithm = initialize_algorithm(config, datasets)  
        
        # load evaluation model
        if config.eval_last:
            eval_model_path = os.path.join(config.save_path_dir, "last_model.pt")
            is_best = False
        else:
            eval_model_path = os.path.join(config.save_path_dir, 'best_model.pt')
            is_best = True

        epoch, best_val_metric = load_algorithm(algorithm, eval_model_path,eval_logger)
        evaluate(algorithm, datasets, config, eval_logger, epoch, is_best)

        eval_logger.close()
    logger.close()

if __name__ == "__main__":
    config = argparser.parse_args()
    main(config)



t5_config = {
    "model": "t5-base",
    "optimizer": "Adam",
    "learning_rate": 3e-5,
    "fp16": False,
    "effective_batch_size":60,
    "max_dialogue_length": 512
}
gpt2_config = {
    "model": "gpt2",
    "optimizer": "Adam",
    "learning_rate": 3e-5,
    "fp16": True,
    "effective_batch_size":60,
    "max_dialogue_length": 512
}
bert_config = {
    "model": "bert-base-uncased",
    "optimizer": "Adam",
    "learning_rate": 3e-5,
    "fp16": True,
    "effective_batch_size":60,
    "max_dialogue_length": 0
}


def initialize_model(config, datasets):
    """
    Initialize models according to the configuration
    """
    if "bert" in config.model:
        model = initialize_bert_based_model(config, datasets)
    elif "gpt" in config.model:
        model = initialize_gpt2_based_model(config)
    elif "t5" in config.model:
        model = initialize_t5_based_model(config)
    return model


def initialize_bert_based_model(config, datasets):
    """
    Initialize BERT based model
    """
    from models.bert import Bert
    model = Bert(config, datasets)
    return model

def initialize_gpt2_based_model(config):
    """
    Initialize GPT2 based model
    """
    from models.gpt2 import GPT2
    model = GPT2(config)
    return model

def initialize_t5_based_model(config):
    """
    Initialize t5 based model
    """
    from models.t5 import T5
    model = T5(config)
    return model


from torch.optim import SGD, Adam
from transformers import AdamW, Adafactor

def initialize_optimizer(config, model):
    # initialize optimizers
    lr = config.learning_rate
    if config.optimizer=='SGD':
        params = model.parameters()
        optimizer = SGD(
            params,
            lr=lr,
            weight_decay=config.weight_decay)
    elif config.optimizer=='AdamW':
        if 'bert' in config.model or 'gpt' in config.model:
            no_decay = ['bias', 'LayerNorm.weight']
        else:
            no_decay = []
        params = [
            {'params': [p for n, p in model.named_parameters() if not any(nd in n for nd in no_decay)], 'weight_decay': config.weight_decay},
            {'params': [p for n, p in model.named_parameters() if any(nd in n for nd in no_decay)], 'weight_decay': 0.0}
            ]
        optimizer = AdamW(
            params,
            lr=lr)
    elif config.optimizer == 'Adam':
        params = model.parameters()
        optimizer = Adam(
            params,
            lr=lr,
            weight_decay=config.weight_decay)
    elif config.optimizer == 'Adafactor':
        params = model.parameters()
        optimizer = Adafactor(
            params,
            scale_parameter=False,
            relative_step=False,
            warmup_init=False,
            lr=lr,
            )
    else:
        raise ValueError(f'Optimizer {config.optimizer} not recognized.')

    return optimizer


from .TLiDB_dataset import TLiDB_Dataset, load_split_ids
import random

class DailyDialog_dataset(TLiDB_Dataset):
    """
    DailyDialog dataset
    This dataset contains all available annotations (currently) from:
        - DailyDialog - http://yanran.li/files/ijcnlp_dailydialog.zip
            - emotion recognition
            - dialog act classification
            - topic classification
        - RECCON - https://github.com/declare-lab/RECCON
            - causal emotion span extraction
            - causal entailment of emotion
        - CIDER - https://github.com/declare-lab/CIDER
            - dialogue level NLI
            - Dialogue reasoning span extraction
            - Dialogue reasoning multiple choice span selection
            - Commonsense relation prediction
        - DailyDialog++ - https://iitmnlp.github.io/DailyDialog-plusplus/
            - Adversarial response selection

    Metadata:
        13118 Total dialogues
        Dialogues per task:
            Emotion recognition: 13118 Dialogues
            Dialog act classification: 13118 Dialogues
            Topic classification: 13118 Dialogues
            Causal emotion span extraction: 1106 Dialogues
            Causal entailment of emotion: 1106 Dialogues
            Dialogue level NLI: 245 Dialogues
            Dialogue reasoning span extraction: 227 Dialogues
            Dialogue reasoning multiple choice span selection: 226 Dialogues
            Commonsense relation prediction: 245 Dialogues
            Adversarial response selection: 6880 Dialogues
    """
    _dataset_name = 'DailyDialog'
    _tasks = [
        'emotion_recognition', 'dialogue_act_classification', 'topic_classification',
        'causal_emotion_span_extraction', 'causal_emotion_entailment',
        'dialogue_nli', 'dialogue_reasoning_span_extraction', 'dialogue_reasoning_multiple_choice_span_selection',
        'dialogue_reasoning_commonsense_relation_prediction', 'adversarial_response_selection',
        'response_generation'
        ]
    _url = "https://drive.google.com/uc?export=download&id=1U9dUi16RbAprUiSBmEKnEpwk45USfnml"
    _task_metadatas = {
        "emotion_recognition": {
                "prompt":"emotion:","type":"classification","loader":"utterance_level_classification",
                "collate_type":"classification", "max_decode_tokens":5
                },
        "dialogue_act_classification": {
            "prompt":"dialogue act:","type":"classification","loader":"utterance_level_classification",
            "collate_type":"classification", "max_decode_tokens":5
            },
        "topic_classification":{
            "prompt":"topic:","type":"classification","loader":"dialogue_level_classification",
            "collate_type":"classification", "max_decode_tokens":5
            },
        "causal_emotion_span_extraction":{# data already contains the prompt
            "prompt":"","type":"span_extraction","loader":"span_extraction",
            "collate_type":"span_extraction", "max_decode_tokens":32
            },
        "causal_emotion_entailment":{
            "prompt":"causal emotion entailment:","type":"classification","loader":"causal_emotion_entailment",
            "collate_type":"nli", "max_decode_tokens":5
            },
        "dialogue_nli":{
            "prompt":"entailment:","type":"classification","loader":"dialogue_nli",
            "collate_type":"nli", "max_decode_tokens":5
        },
        "dialogue_reasoning_span_extraction":{
            "prompt":"", "type":"span_extraction","loader":"span_extraction",
            "collate_type":"span_extraction", "max_decode_tokens":32
        },
        "dialogue_reasoning_multiple_choice_span_selection":{
            "prompt":"The correct option is", "type":"multiple_choice","loader":"multiple_choice",
            "collate_type":"multiple_choice", "num_choices":4, "max_decode_tokens":20
        },
        "dialogue_reasoning_commonsense_relation_prediction":{
            "prompt":"", "type":"classification","loader":"relation_extraction",
            "collate_type":"relation_extraction", "max_decode_tokens":5
        },
        "adversarial_response_selection":{
            "prompt":"The correct option is","type":"multiple_choice","loader":"adversarial_response_selection",
            "collate_type":"multiple_choice", "num_choices":3, "max_decode_tokens":1
        },
        "response_generation":{
            "prompt":"", "type":"response_generation","loader":"response_generation",
            "max_decode_tokens":64
        }
    }
    def __init__(self, task, dataset_folder, model_type, split, max_dialogue_length, few_shot_percent=None):
        assert task in self._tasks, f"{task} is not a valid task for {self._dataset_name}"
        super().__init__(self._dataset_name, task, model_type, max_dialogue_length, dataset_folder=dataset_folder)
        self._task_metadata = self._task_metadatas[task]
        self._input_array = []
        self._y_array = []
        self._metadata_fields = []
        self._metadata_array = []
        split_ids = load_split_ids(self._dataset_name, dataset_folder, split, few_shot_percent)
        self._load_data(task, split_ids)
        self._num_classes = len(self.task_labels)
        self._y_size = len(self._y_array)
        

    def _load_data(self, task, split_ids):
        # get the data loader, based on whether the task is utterance level/dialogue level/span extraction/etc.
        loader = getattr(self, f"_load_{self._task_metadata['loader']}_task")
        return loader(task,split_ids)

    def _load_utterance_level_classification_task(self, task, split_ids):
        for datum in self.dataset['data']:
            if datum['dialogue_id'] in split_ids:
                if task in datum['dialogue_metadata']:
                    dialogue = []
                    for turn in datum['dialogue']:
                        dialogue.append([turn['speakers'][0], turn['utterance']])
                        str_dialogue = self._convert_dialogue_to_string(dialogue)
                        if task in turn:
                            self._input_array.append(str_dialogue)
                            self._y_array.append(turn[task])

    def _load_dialogue_level_classification_task(self, task, split_ids):
        for datum in self.dataset['data']:
            if datum['dialogue_id'] in split_ids:
                if task in datum['dialogue_metadata']:            
                    dialogue = [[turn['speakers'][0], turn['utterance']] for turn in datum['dialogue']]
                    str_dialogue = self._convert_dialogue_to_string(dialogue)
                    self._input_array.append(str_dialogue)
                    self._y_array.append(datum[task])
        
    def _load_span_extraction_task(self, task, split_ids):
        for datum in self.dataset['data']:
            if datum['dialogue_id'] in split_ids:
                if task in datum['dialogue_metadata']:
                    for qas in datum[task]:
                        for qa in qas['qas']:
                            for answer in qa['answers']:
                                self._input_array.append({
                                    "context":qas['context'],"question":qa['question']
                                })
                                # move the answer start back by len(question) if not impossible
                                if answer['answer_start'] > 0:
                                    answer['answer_start'] += len(qa['question'])+1
                                self._y_array.append(answer)

    def _load_causal_emotion_entailment_task(self, task, split_ids):
        for datum in self.dataset['data']:
            if datum['dialogue_id'] in split_ids:
                if task in datum['dialogue_metadata']:
                    for sample in datum[task]:
                        self._input_array.append({
                            "premise":sample['history'],
                            "hypothesis": f"{sample['causal_utterance']} causes {sample['emotion']} in {sample['target_utterance']}",
                        })
                        self._y_array.append(sample['labels'])

    def _load_dialogue_nli_task(self, task, split_ids):
        for datum in self.dataset['data']:
            if datum['dialogue_id'] in split_ids:
                if task in datum['dialogue_metadata']:
                    for sample in datum[task]:
                        self._input_array.append({
                            "premise": self._convert_dialogue_to_string([[turn['speakers'][0], turn['utterance']] for turn in datum['dialogue']]),
                            "hypothesis": f"{sample['head']} {sample['relation']} {sample['tail']}"
                        })
                        self._y_array.append(sample['label'])

    def _load_multiple_choice_task(self, task, split_ids):
        for datum in self.dataset['data']:
            if datum['dialogue_id'] in split_ids:
                if task in datum['dialogue_metadata']:
                    context = datum[task]['context']
                    mcqs = datum[task]['mcqs']
                    for q in mcqs:
                        self._input_array.append({
                            "context": context,
                            "question": q['question'],
                            "options": q['options']
                        })
                        self._y_array.append(q['label'])

    def _load_relation_extraction_task(self, task, split_ids):
        for datum in self.dataset['data']:
            if datum['dialogue_id'] in split_ids:
                if task in datum['dialogue_metadata']:
                    dialogue = [[turn['speakers'][0], turn['utterance']] for turn in datum['dialogue']]
                    dialogue = self._convert_dialogue_to_string(dialogue)
                    for sample in datum[task]:
                        self._input_array.append({
                            "context": dialogue,
                            "head": sample['head'],
                            "tail": sample['tail']
                        })
                        self._y_array.append(sample['relation'])
                
    def _load_adversarial_response_selection_task(self, task, split_ids):
        for datum in self.dataset['data']:
            if datum['dialogue_id'] in split_ids:
                if task in datum['dialogue_metadata']:
                    for sample in datum[task]:
                        context = []
                        for turn in datum['dialogue']:
                            if turn['turn_id'] in sample['context_turns']:
                                context.append([turn['speakers'][0], turn['utterance']])
                        context = self._convert_dialogue_to_string(context)
                        for pos_resp, random_neg_resp, adv_neg_resp in zip(sample['positive_responses'], sample['random_negative_responses'], sample['adversarial_negative_responses']):
                            # shuffle the options
                            options = [pos_resp, random_neg_resp, adv_neg_resp]
                            random.shuffle(options)
                            self._input_array.append({
                                "context": context,
                                "options": options,
                                "question":"Which option is the best response?"
                            })
                            self._y_array.append(str(options.index(pos_resp)))

    def _collate_encoder(self, batch):
        X, y, metadata = [], [], {}
        for item in batch:
            if self._task_metadata['collate_type'] == 'span_extraction':
                X.append(self._join_strings(item[0]['question'],item[0]['context']))
            elif self._task_metadata['collate_type'] == 'nli':
                X.append(self._join_strings(item[0]['premise'], item[0]['hypothesis']))
            elif self._task_metadata['collate_type'] == 'classification':
                X.append(item[0])
            elif self._task_metadata['collate_type'] == 'multiple_choice':
                mcq_inputs = []
                context = item[0]['context']
                question = item[0]['question']
                options = item[0]['options']
                for option in options:
                    mcq_inputs.append(self._join_strings(context,"[SEP]", question, "[SEP]", option))
                X.append(mcq_inputs)
            elif self._task_metadata['collate_type'] == "relation_extraction":
                X.append(self._join_strings(item[0]['head'],"[SEP]",item[0]['tail'],"[SEP]",item[0]['context']))
            else:
                raise NotImplementedError(f"Collate type {self._task_metadata['collate_type']} not implemented")
            y.append(item[1])
            for k, v in item[2].items():
                if k not in metadata:
                    metadata.append(k)
                metadata[k].append(v)
        return X, y, metadata

    def _collate_decoder(self, batch):
        X, y, metadata = [], [], {}
        for item in batch:
            if self._task_metadata['collate_type'] == 'span_extraction':
                X.append(self._join_strings("context:",item[0]['context'],
                                            "question:", item[0]['question'], "answer:"))
            elif self._task_metadata['collate_type'] == 'nli':
                X.append(self._join_strings("context:",item[0]['premise'],
                                            self._task_metadata['prompt'],item[0]['hypothesis']))
            elif self._task_metadata['collate_type'] == 'classification':
                dialogue = item[0]
                X.append(self._join_strings("context:",dialogue,self._task_metadata['prompt']))
            elif self._task_metadata['collate_type'] == 'multiple_choice':
                options_str = " ".join(f"option {i}: {option}" for i, option in enumerate(item[0]['options']))
                X.append(self._join_strings("context:",item[0]['context'],
                                            "question:",item[0]['question'],\
                                            options_str,self._task_metadata['prompt']))
            elif self._task_metadata['collate_type'] == "relation_extraction":
                X.append(self._join_strings("context:",item[0]['context'],
                                            f"The relation between '{item[0]['head']}' and '{item[0]['tail']}' is"))
            else:
                raise NotImplementedError(f"Collate type {self._task_metadata['collate_type']} not implemented")       
            y.append(item[1])
            for k, v in item[2].items():
                if k not in metadata:
                    metadata.append(k)
                metadata[k].append(v)
        labels = self.task_labels
        if labels:
            metadata['labels'] = labels
        return X, y, metadata

    def _collate_encoderdecoder(self, batch):
        X, y, metadata = [], [], {}
        for item in batch:
            if self._task_metadata['collate_type'] == 'span_extraction':
                X.append(self._join_strings("context:",item[0]['context'],
                                            "question:", item[0]['question'], "answer:"))
            elif self._task_metadata['collate_type'] == 'nli':
                X.append(self._join_strings("context:",item[0]['premise'],
                                            self._task_metadata['prompt'],item[0]['hypothesis']))
            elif self._task_metadata['collate_type'] == 'classification':
                dialogue = item[0]
                X.append(self._join_strings("context:",dialogue,self._task_metadata['prompt']))
            elif self._task_metadata['collate_type'] == 'multiple_choice':
                options_str = " ".join(f"option {i}: {option}" for i, option in enumerate(item[0]['options']))
                X.append(self._join_strings("context:",item[0]['context'],
                                            "question:",item[0]['question'],\
                                            options_str,self._task_metadata['prompt']))
            elif self._task_metadata['collate_type'] == "relation_extraction":
                X.append(self._join_strings("context:",item[0]['context'],
                                            f"The relation between '{item[0]['head']}' and '{item[0]['tail']}' is"))
            else:
                raise NotImplementedError(f"Collate type {self._task_metadata['collate_type']} not implemented")       
            y.append(item[1])
            for k, v in item[2].items():
                if k not in metadata:
                    metadata.append(k)
                metadata[k].append(v)
        labels = self.task_labels
        if labels:
            metadata['labels'] = labels
        return X, y, metadata


from algorithms.EncoderAlgorithm import EncoderAlgorithm
from examples.algorithms.EncoderDecoderAlgorithm import EncoderDecoderAlgorithm
from algorithms.DecoderAlgorithm import DecoderAlgorithm

def initialize_algorithm(config, datasets):
    """Load an algorithm of type Algorithm
    Args:
        config (dict): configuration dictionary
        datasets (dict): dictionary of datasets
    Returns:    
        algorithm (Algorithm): an algorithm object
    """
    if config.model_type == "Encoder":
        algorithm = EncoderAlgorithm(config, datasets)
    elif config.model_type == "Decoder":
        algorithm = DecoderAlgorithm(config, datasets)
    elif config.model_type == "EncoderDecoder":
        algorithm = EncoderDecoderAlgorithm(config, datasets)
    else:
        raise ValueError(f"Invalid algorithm name: {config.model_type}")

    return algorithm


from transformers import T5ForConditionalGeneration, T5Tokenizer
from .TLiDB_model import TLiDB_model
import random

class T5(TLiDB_model):
    def __init__(self, config):
        super().__init__(config)
        self.tokenizer, self.model = initialize_model(config)
        self.layers = {"model":self.model}
        self.init_weights()

    def load_state_dict(self, state_dict):
        self.model.load_state_dict(state_dict['model'])

    def _forward(
        self,
        input_ids = None,
        attention_mask = None,
        encoder_outputs = None,
        decoder_input_ids = None,
        decoder_attention_mask = None,
        lm_labels = None,
    ):

        outputs = self.model(input_ids=input_ids,
                             attention_mask=attention_mask,
                             decoder_input_ids=decoder_input_ids,
                             decoder_attention_mask=decoder_attention_mask,
                             encoder_outputs=encoder_outputs,
                             labels=lm_labels)
        return outputs.loss

    def transform_inputs(self, inputs):
        """Only tokenizes inputs"""
        tokenized_inputs = self.tokenizer(inputs, padding="longest", pad_to_multiple_of=8,
                                        truncation=True, return_tensors="pt")
        return tokenized_inputs

    def transform_outputs(self, outputs):
        """tokenizes outputs"""
        # if multiple correct outputs, select 1 at random
        if isinstance(outputs[0], list):
            outputs = [random.choice(output) for output in outputs]
        tokenized_outputs = self.tokenizer(outputs, padding="longest", pad_to_multiple_of=8,
                                        truncation=True, return_tensors="pt")
        # replace pad tokens by -100
        label_ids = tokenized_outputs.input_ids
        label_ids[label_ids == self.tokenizer.pad_token_id] = -100
        return label_ids

    def generate(self, input_ids, **kwargs):
        pred_tokens = self.model.generate(input_ids=input_ids, **kwargs)
        return pred_tokens

    def batch_decode(self, tokens):
        return self.tokenizer.batch_decode(tokens, skip_special_tokens=True)

def initialize_model(config):
    tokenizer = T5Tokenizer.from_pretrained(config.model)
    model = T5ForConditionalGeneration.from_pretrained(config.model)
    model.resize_token_embeddings(len(tokenizer))
    return tokenizer, model



import torch
from .TLiDB_model import TLiDB_model
from examples.utils import concat_t_d
import random

class Bert(TLiDB_model):
    def __init__(self, config, datasets):
        super().__init__(config)
        self.tokenizer, self.model = initialize_model(config)
        self.dropout = torch.nn.Dropout(self.model.config.hidden_dropout_prob)
        self.layers = {"model":self.model}
        # for each task/domain, we add a new classification layer to the model
        self.classifiers = {}
        for split in datasets.keys():
            for d in datasets[split]['datasets']:
                # make this more flexible, call a layer initializer which knows what task to create the layer for
                t_d = concat_t_d(d.task,d.dataset_name)
                if t_d not in self.classifiers.keys():
                    task_type = d.task_metadata['type']
                    layer = self.initialize_bert_classifier(task_type, d)
                    setattr(self, f"{t_d}_classifier", layer)
                    forward = self.initialize_forward(task_type)
                    self.classifiers[t_d] = {
                        "classifier": getattr(self, f"{t_d}_classifier"),
                        "labels":d.task_labels,
                        "forward":forward}
                    self.layers[f"{t_d}_classifier"] = getattr(self, f"{t_d}_classifier")

        self.init_weights()

    def initialize_bert_classifier(self, task_type, dataset):
        if task_type in ["classification","multilabel_classification"]:
        # dataset.task in SEQUENCE_TASKS:
            return torch.nn.Linear(self.model.config.hidden_size, dataset.num_classes)
        elif task_type == "multioutput_classification":
        # elif dataset.task in MULTIOUTPUT_SEQUENCE_TASKS:
            return torch.nn.Linear(self.model.config.hidden_size, dataset.num_classes*dataset.task_metadata['num_outputs'])
        elif task_type == "token_classification":
        # elif dataset.task in TOKEN_TASKS:
            return torch.nn.Linear(self.model.config.hidden_size, dataset.num_classes)
        elif task_type == "span_extraction":
        # elif dataset.task in SPAN_EXTRACTION_TASKS:
            return torch.nn.Linear(self.model.config.hidden_size, 2)
        elif task_type == "multiple_choice":
        # elif dataset.task in MULTIPLE_CHOICE_TASKS:
            return torch.nn.Linear(self.model.config.hidden_size, 1)
        else:
            raise ValueError(f"Unsupported task: {dataset.task}")

    def initialize_forward(self, task_type):
        if task_type in ['classification', 'multioutput_classification','multilabel_classification']:
            return self.sequence_classification
        elif task_type == 'token_classification':
            return self.token_classification
        elif task_type == 'span_extraction':
            return self.span_extraction
        elif task_type == 'multiple_choice':
            return self.multiple_choice
        else:
            raise ValueError(f"Unsupported task: {task_type}")

    def load_state_dict(self, state_dict):
        for layer_name, layer in state_dict.items():
            if layer_name in self.layers.keys():
                self.layers[layer_name].load_state_dict(layer)

    def _forward(self, inputs, task, dataset_name):
        return self.classifiers[concat_t_d(task,dataset_name)]['forward'](inputs, task, dataset_name)

    def transform_inputs(self, inputs, metadata):
        """Only tokenizes inputs"""
        # specify if we need the offset mapping
        if 'return_offsets_mapping' in metadata:
            return_offsets_mapping = metadata['return_offsets_mapping']
        else:
            return_offsets_mapping = False

        tokenized_inputs = self.tokenizer(inputs, padding="longest",pad_to_multiple_of=8,truncation=True,
                                        return_offsets_mapping=return_offsets_mapping,return_tensors="pt")
        return tokenized_inputs
    
    def transform_outputs(self, inputs, outputs, task_type, metadata):
        """Calls the classification layer associated with task and dataset_name"""
        t_d = concat_t_d(metadata['task'],metadata['dataset_name'])
        outputs = getattr(self, f"transform_{task_type}_outputs")(inputs, outputs, t_d, metadata)
        return outputs

    def transform_classification_outputs(self,inputs, outputs, t_d, metadata):
        outputs = [self.classifiers[t_d]['labels'].index(y) for y in outputs]
        return torch.tensor(outputs, dtype=torch.long)

    def transform_multioutput_classification_outputs(self, inputs, outputs, t_d, metadata):
        outputs = [[self.classifiers[t_d]['labels'].index(y) for y in output.values()] for output in outputs]
        return torch.tensor(outputs, dtype=torch.long)

    def transform_multilabel_classification_outputs(self, inputs, outputs, t_d, metadata):
        converted_outputs = []
        for output in outputs:
            converted_output = [0 for _ in range(len(self.classifiers[t_d]['labels']))]
            for o in output:
                converted_output[self.classifiers[t_d]['labels'].index(o)] = 1
            converted_outputs.append(converted_output)
        return torch.tensor(converted_outputs, dtype=torch.float)

    def transform_span_extraction_outputs(self, inputs, outputs, t_d, metadata):
        start_indices, end_indices = [], []

        for offset_mapping,output in zip(inputs.offset_mapping,outputs):
            # if multiple correct spans, randomly sample
            if isinstance(output, list):
                answer_index = random.randint(0, len(output)-1)
                text = output[answer_index]['text']
                answer_start = output[answer_index]['answer_start']
            else:
                text = output['text']
                answer_start = output['answer_start']

            start_idx,end_idx = get_token_offsets(offset_mapping, text, answer_start)
            start_indices.append(start_idx)
            end_indices.append(end_idx)
        
        return [torch.tensor(start_indices,dtype=torch.long), torch.tensor(end_indices,dtype=torch.long)]
    
    def transform_multiple_choice_outputs(self, inputs, outputs, t_d, metadata):
        return torch.tensor(outputs, dtype=torch.long)

    # classify a sequence
    def sequence_classification(self, tokenized_sequences, task, dataset_name):
        """Classify a sequence of tokens with a single label"""
        t_d = concat_t_d(task,dataset_name)
        outputs = self.model(input_ids=tokenized_sequences.input_ids, attention_mask=tokenized_sequences.attention_mask)['pooler_output']
        outputs = self.dropout(outputs)
        logits = self.classifiers[t_d]['classifier'](outputs)
        return logits

    # classify a sequence of tokens
    def token_classification(self, tokenized_sequences, task, dataset_name):
        """Classify each token in a sequence"""
        t_d = concat_t_d(task,dataset_name)
        outputs = self.model(input_ids=tokenized_sequences.input_ids, attention_mask=tokenized_sequences.attention_mask)['last_hidden_state']
        outputs = self.dropout(outputs)
        logits = self.classifiers[t_d]['classifier'](outputs)
        return logits

    # extract spans
    def span_extraction(self, tokenized_sequences, task, dataset_name):
        """Extract spans from a sequence of tokens"""
        t_d = concat_t_d(task,dataset_name)
        outputs = self.model(input_ids=tokenized_sequences.input_ids, attention_mask=tokenized_sequences.attention_mask)['last_hidden_state']
        outputs = self.dropout(outputs)
        logits = self.classifiers[t_d]['classifier'](outputs)
        return logits

    def multiple_choice(self, tokenized_sequences, task, dataset_name):
        """Multiple choice classification"""
        t_d = concat_t_d(task,dataset_name)
        outputs = self.model(input_ids=tokenized_sequences.input_ids, attention_mask=tokenized_sequences.attention_mask)['pooler_output']
        outputs = self.dropout(outputs)
        logits = self.classifiers[t_d]['classifier'](outputs)
        return logits

def initialize_model(config):
    if config.model in ["bert-base-uncased"]:
        from transformers import BertModel, BertTokenizerFast
        tokenizer = BertTokenizerFast.from_pretrained(config.model)
        model = BertModel.from_pretrained(config.model)
        return tokenizer, model
    else:
        raise ValueError(f"Unsupported BERT model: {config.model}")

def get_token_offsets(offsets_mapping, text, text_start_idx):
    """
    Given a start index and length of the text, return the start and end indices of the span
    """
    if text_start_idx < 0:
        return 0,0
    token_start_idx, token_end_idx = 0,0
    text_end_index = text_start_idx + len(text)
    for i, (start, end) in enumerate(offsets_mapping):
        if not token_start_idx and start <= text_start_idx and end >= text_start_idx:
            token_start_idx = i
        if not token_end_idx and start <= text_end_index and end >= text_end_index:
            token_end_idx = i
    return token_start_idx, token_end_idx


import re
import string
import numpy as np
import torch
from TLiDB.utils.utils import numel

class Metric:
    """
    Parent class for metrics
    """
    def __init__(self,name):
        self._name = name

    def _compute(self, y_pred, y_true):
        """
        Helper function for computing the metric.
        Subclasses should implement this.
        Args:
            - y_pred (Tensor): Predicted targets or model output
            - y_true (Tensor): True targets
        Output:
            - metric (0-dim tensor): metric
        """
        return NotImplementedError

    @property
    def name(self):
        """
        Metric name.
        Used to name the key in the results dictionaries returned by the metric.
        """
        return self._name

    @property
    def agg_metric_field(self):
        """
        The name of the key in the results dictionary returned by Metric.compute().
        This should correspond to the aggregate metric computed on all of y_pred and y_true,
        in contrast to a group-wise evaluation.
        """
        return f'{self.name}_all'

    def compute(self, y_pred, y_true, return_dict=True):
        """
        Computes metric. This is a wrapper around _compute.
        Args:
            - y_pred (Tensor): Predicted targets or model output
            - y_true (Tensor): True targets
            - return_dict (bool): Whether to return the output as a dictionary or a tensor
        Output (return_dict=False):
            - metric (0-dim tensor): metric. If the inputs are empty, returns tensor(0.)
        Output (return_dict=True):
            - results (dict): Dictionary of results, mapping metric.agg_metric_field to avg_metric
        """
        if numel(y_true) == 0:
            agg_metric = torch.tensor(0., device=y_true.device)
        else:
            agg_metric = self._compute(y_pred, y_true)
        if return_dict:
            results = {
                self.agg_metric_field: agg_metric.item()
            }
            return results
        else:
            return agg_metric

class StringMetric:
    """
    Parent class for string metrics
    """
    def __init__(self,name,unanswerable_phrases, ignore_unanswerable=False):
        self._name = name
        self._unanswerable_phrases = [self._normalize_answer(text, string.punctuation, '') for text in unanswerable_phrases]
        self._ignore_unanswerable = ignore_unanswerable

    def _compute(self, y_pred, y_true):
        """
        Helper function for computing the metric.
        Subclasses should implement this.
        Args:
            - y_pred (List of str): Predicted targets or model output
            - y_true (List of str OR List of List of str): True targets
        Output:
            - metric (0-dim tensor): metric
        """
        return NotImplementedError

    @property
    def name(self):
        """
        Metric name.
        Used to name the key in the results dictionaries returned by the metric.
        """
        return self._name

    @property
    def unanswerable_phrases(self):
        """
        List of phrases to ignore when computing the metric.
        """
        return self._unanswerable_phrases

    @property
    def agg_metric_field(self):
        """
        The name of the key in the results dictionary returned by Metric.compute().
        This should correspond to the aggregate metric computed on all of y_pred and y_true,
        in contrast to a group-wise evaluation.
        """
        return f'{self.name}_all'

    def compute(self, y_pred, y_true, return_dict=True):
        """
        Computes metric. This is a wrapper around _compute.
        Args:
            - y_pred (List of str): Predicted targets or model output
            - y_true (List of str OR List of List of str): True targets
            - return_dict (bool): Whether to return the output as a dictionary or a tensor
        Output (return_dict=False):
            - metric (0-dim tensor): metric. If the inputs are empty, returns tensor(0.)
        Output (return_dict=True):
            - results (dict): Dictionary of results, mapping metric.agg_metric_field to avg_metric
        """
        if numel(y_true) == 0:
            agg_metric = torch.tensor(0., device=y_true.device)
        else:
            y_pred = [self._normalize_answer(text, string.punctuation, '') for text in y_pred]

            if isinstance(y_true[0], list):
                y_true = [[self._normalize_answer(text, string.punctuation, '') for text in answers] for answers in y_true]
            else:
                y_true = [self._normalize_answer(text, string.punctuation, '') for text in y_true]

            if self._unanswerable_phrases:
                y_pred = [text if not any([unanswerable_phrase in text for unanswerable_phrase in self.unanswerable_phrases]) else "" for text in y_pred]
                
                if isinstance(y_true[0], list):
                    y_true = [[text if not any([unanswerable_phrase in text for unanswerable_phrase in self.unanswerable_phrases]) else "" for text in answers] for answers in y_true]
                else:
                    y_true = [text if not any([unanswerable_phrase in text for unanswerable_phrase in self.unanswerable_phrases]) else "" for text in y_true]
                
                if self._ignore_unanswerable:
                    pos_pred, pos_true = [], []
                    for pred, true in zip(y_pred, y_true):
                        if true != '':
                            pos_pred.append(pred)
                            pos_true.append(true)
                    y_pred = pos_pred
                    y_true = pos_true

            agg_metric = self._compute(y_pred, y_true)
        if return_dict:
            results = {
                self.agg_metric_field: agg_metric
            }
            return results
        else:
            return agg_metric

    def _normalize_answer(self, text, punc_chars, punc_repl):
        """
        Lower text and remove punctuation, articles and extra whitespace.
        Shamelessly copied from https://github.com/google-research/text-to-text-transfer-transformer/blob/220e43384912392302c34aaea9398dc5d66d975b/t5/evaluation/qa_utils.py#L29
        """

        def remove_articles(s):
            return re.sub(r"\b(a|an|the)\b", " ", s)

        def replace_punctuation(s):
            to_replace = set(punc_chars)
            return "".join(punc_repl if ch in to_replace else ch for ch in s)

        def white_space_fix(s):
            return " ".join(s.split())

        text = text.lower()
        text = replace_punctuation(text)
        text = remove_articles(text)
        text = white_space_fix(text)

        return text

    def _metric_max_over_ground_truths(self, metric_fn, prediction, ground_truths):
        """
        Computes the maximum of the metric over all ground truths.
        Shamelessly copied from https://github.com/google-research/text-to-text-transfer-transformer/blob/cec7078ac27a2d98750279c158d25d9d1df16b3a/t5/evaluation/qa_utils.py#L61
        """
        return max(
            metric_fn(prediction, ground_truth) for ground_truth in ground_truths
        )

class ElementwiseMetric(Metric):
    """
    Averages.
    """
    def _compute_element_wise(self, y_pred, y_true):
        """
        Helper for computing element-wise metric, implemented for each metric
        Args:
            - y_pred (Tensor): Predicted targets or model output
            - y_true (Tensor): True targets
        Output:
            - element_wise_metrics (Tensor): tensor of size (batch_size, )
        """
        raise NotImplementedError

    def _compute(self, y_pred, y_true):
        """
        Helper function for computing the metric.
        Args:
            - y_pred (Tensor): Predicted targets or model output
            - y_true (Tensor): True targets
        Output:
            - avg_metric (0-dim tensor): average of element-wise metrics
        """
        element_wise_metrics = self._compute_element_wise(y_pred, y_true)
        avg_metric = element_wise_metrics.mean()
        return avg_metric

    @property
    def agg_metric_field(self):
        """
        The name of the key in the results dictionary returned by Metric.compute().
        """
        return f'{self.name}_avg'

    def compute_element_wise(self, y_pred, y_true, return_dict=True):
        """
        Computes element-wise metric
        Args:
            - y_pred (Tensor): Predicted targets or model output
            - y_true (Tensor): True targets
            - return_dict (bool): Whether to return the output as a dictionary or a tensor
        Output (return_dict=False):
            - element_wise_metrics (Tensor): tensor of size (batch_size, )
        Output (return_dict=True):
            - results (dict): Dictionary of results, mapping metric.name to element_wise_metrics
        """
        element_wise_metrics = self._compute_element_wise(y_pred, y_true)
        batch_size = y_pred.size()[0]
        assert element_wise_metrics.dim() == 1 and element_wise_metrics.numel() == batch_size

        if return_dict:
            return {self.name: element_wise_metrics}
        else:
            return element_wise_metrics



import random
from .TLiDB_model import TLiDB_model

class GPT2(TLiDB_model):
    def __init__(self, config):
        super().__init__(config)
        self.tokenizer, self.model = initialize_model(config)
        self.layers = {"model":self.model}
    
    def load_state_dict(self, state_dict):
        self.model.load_state_dict(state_dict['model'])

    def _forward(
        self,
        input_ids = None,
        attention_mask = None,
        lm_labels = None
    ):
        outputs = self.model(
            input_ids=input_ids,
            attention_mask=attention_mask,
            labels=lm_labels)
        return outputs.loss


    def transform_LM_inputs(self, inputs, outputs):
        # if multiple correct outputs, select 1 at random
        if isinstance(outputs[0], list):
            outputs = [random.choice(output) for output in outputs]
        tokenized_inputs = self.tokenizer([" ".join([i,o])+self.tokenizer.eos_token for i,o in zip(inputs,outputs)],
                                        padding="longest",pad_to_multiple_of=8,truncation=True, return_tensors="pt")
        labels = tokenized_inputs.input_ids.detach().clone()
        # replace pad tokens by -100
        labels[labels == self.tokenizer.pad_token_id] = -100
        return tokenized_inputs, labels

    def transform_generation_inputs(self, inputs):
        self.tokenizer.padding_side = "left"
        tokenized_inputs = self.tokenizer(inputs, padding="longest", pad_to_multiple_of=8,
                                        truncation=True, return_tensors="pt")
        self.tokenizer.padding_side = "right"
        return tokenized_inputs


    def generate(self, X, max_decode_tokens, **kwargs):
        input_size = X.input_ids.shape[-1]
        outputs = self.model.generate(input_ids=X.input_ids,attention_mask=X.attention_mask,
                                    max_length=input_size+max_decode_tokens, **kwargs)
        pred_tokens = outputs[:, input_size:]
        preds = self.tokenizer.batch_decode(pred_tokens, skip_special_tokens=True)
        preds = [pred.strip() for pred in preds]
        return preds


def initialize_model(config):
    if 'neo' in config.model:
        from transformers import GPTNeoForCausalLM, GPT2Tokenizer
        tokenizer = GPT2Tokenizer.from_pretrained(config.model, pad_token='<|pad|>')
        model = GPTNeoForCausalLM.from_pretrained(config.model)
        return tokenizer, model
    else:
        from transformers import GPT2LMHeadModel, GPT2TokenizerFast
        tokenizer = GPT2TokenizerFast.from_pretrained(config.model, pad_token='<|pad|>')
        model = GPT2LMHeadModel.from_pretrained(config.model)
        model.config.pad_token_id = tokenizer.pad_token_id
        model.resize_token_embeddings(len(tokenizer))
        return tokenizer, model



import os
import sys
import random
import numpy as np
import torch

from TLiDB.datasets.get_dataset import get_dataset
from TLiDB.data_loaders.data_loaders import get_loader
from TLiDB.metrics.initializer import get_metric_computer

def load_datasets_split(split, tasks, datasets, config):
    split_datasets = {"datasets":[], "loaders":[], "metrics":[]}
    get_data_loader = get_loader(split)
    for t, d in zip(tasks, datasets):
        cur_dataset = get_dataset(dataset=d,task=t,dataset_folder=config.data_dir,
                            model_type=config.model_type,
                            max_dialogue_length=config.max_dialogue_length,
                            split=split,few_shot_percent=config.few_shot_percent)
        if config.frac < 1.0:
            cur_dataset.random_subsample(config.frac)

        split_datasets["datasets"].append(cur_dataset)
        split_datasets["loaders"].append(get_data_loader(cur_dataset, config.gpu_batch_size, config, collate_fn=cur_dataset.collate, num_workers=config.num_workers))
        split_datasets["metrics"].append(get_metric_computer(cur_dataset.metrics, **cur_dataset.metric_kwargs))
    return split_datasets

def concat_t_d(task, dataset_name):
    return f"{task}_{dataset_name}"

def move_to(obj, device):
    if isinstance(obj, dict):
        return {k: move_to(v, device) for k, v in obj.items()}
    elif isinstance(obj, list):
        return [move_to(v, device) for v in obj]
    elif isinstance(obj, float) or isinstance(obj, int) or isinstance(obj, str) or obj is None:
        return obj
    else:
        # Assume obj is a Tensor or other type
        return obj.to(device)

def detach_and_clone(obj):
    if torch.is_tensor(obj):
        return obj.detach().clone()
    elif isinstance(obj, dict):
        return {k: detach_and_clone(v) for k, v in obj.items()}
    elif isinstance(obj, list):
        return [detach_and_clone(v) for v in obj]
    elif isinstance(obj, float) or isinstance(obj, int) or isinstance(obj, str) or obj is None:
        return obj
    else:
        raise TypeError("Invalid type for detach_and_clone")


def collate_list(inputs):
    """
    If inputs is a list of Tensors, it concatenates them all along the first dimension.

    If inputs is a list of lists, it joins these lists together, but does not attempt to
    recursively collate. This allows each element of the list to be, e.g., its own dict.

    If inputs is a list of dicts (with the same keys in each dict), it returns a single dict
    with the same keys. For each key, it recursively collates all entries in the list.
    """
    if not isinstance(inputs, list):
        raise TypeError("collate_list must take in a list")
    elem = inputs[0]
    if torch.is_tensor(elem):
        return torch.cat(inputs)
    elif isinstance(elem, list):
        return [obj for sublist in inputs for obj in sublist]
    elif isinstance(elem, dict):
        return {k: collate_list([d[k] for d in inputs]) for k in elem}
    else:
        raise TypeError("Elements of the list to collate must be tensors or dicts.")


def set_seed(seed):
    if seed != -1:
        random.seed(seed)
        np.random.seed(seed)
        torch.manual_seed(seed)

def get_savepath_dir(datasets, tasks, seed, log_dir, model, few_shot_percent, cotraining=False):
    prefix = "PRETRAINED_"
    if cotraining:
        prefix = "COTRAINED_"
    if few_shot_percent:
        prefix += f"{few_shot_percent}_FEWSHOT_"
    for dataset,task in zip(datasets, tasks):
        prefix += f"{dataset}.{task}_"
    if seed > -1:
        prefix += f"seed.{seed}_"
    if prefix == "PRETRAINED_":
        raise ValueError("Cannot create dir with empty name")

    prefix = os.path.join(log_dir, prefix[:-1], model)
    return prefix

def append_to_save_path_dir(save_path_dir, datasets, tasks, few_shot_percent, seed):
    postfix = "FINETUNED_"
    if few_shot_percent:
        postfix += f"{few_shot_percent}_FEWSHOT_"
    for dataset,task in zip(datasets, tasks):
        postfix += f"{dataset}.{task}_"
    if seed > -1:
        postfix += f"seed.{seed}_"
    if postfix == "FINETUNED_":
        raise ValueError("Cannot create dir with empty name")
    return os.path.join(save_path_dir,postfix[:-1])

def save_algorithm(algorithm, epoch, best_val_metric, path, logger):
    state = {}
    state['algorithm'] = algorithm.state_dict()
    state['epoch'] = epoch
    state['best_val_metric'] = best_val_metric
    torch.save(state, path)
    logger.write(f"Saved model to {path}\n")

def load_algorithm(algorithm, path, logger):
    state = torch.load(path)
    algorithm.load_state_dict(state['algorithm'])
    logger.write(f"Loaded model from {path}\n")
    return state['epoch'], state['best_val_metric']

def save_algorithm_if_needed(algorithm, epoch, config, best_val_metric, is_best, logger):
    if config.save_last:
        save_algorithm(algorithm,epoch,best_val_metric,os.path.join(config.save_path_dir,"last_model.pt"),logger)
    if config.save_best and is_best:
        save_algorithm(algorithm, epoch, best_val_metric,os.path.join(config.save_path_dir,"best_model.pt"),logger)

def save_pred_if_needed(y_pred, epoch, config, is_best, save_path_dir):
    if config.save_pred:
        if config.save_last:
            save_pred(y_pred, os.path.join(save_path_dir, "last_predictions"))
        if config.save_best and is_best:
            save_pred(y_pred, os.path.join(save_path_dir, "best_predictions"))

def save_pred(y_pred, path_prefix):
    # Single tensor
    if torch.is_tensor(y_pred):
        y_pred_np = y_pred.numpy()
        np.save(path_prefix+'.csv', y_pred_np)
    # Dictionary
    elif isinstance(y_pred, dict) or isinstance(y_pred, list):
        torch.save(y_pred, path_prefix + '.pt')
    else:
        raise TypeError("Invalid type for save_pred")

class Logger(object):
    def __init__(self, fpath=None, mode='w'):
        self.console = sys.stdout
        self.file = None
        if fpath is not None:
            self.file = open(fpath, mode)

    def __del__(self):
        self.close()

    def __enter__(self):
        pass

    def __exit__(self, *args):
        self.close()

    def write(self, msg):
        self.console.write(msg)
        if self.file is not None:
            self.file.write(msg)

    def flush(self):
        self.console.flush()
        if self.file is not None:
            self.file.flush()
            os.fsync(self.file.fileno())

    def close(self):
        # self.console.close()
        if self.file is not None:
            self.file.close()


def log_config(config, logger):
    # Simply write all configuration parameters to the logger
    logger.write('Configuration:\n')
    for name, val in vars(config).items():
        logger.write(f'{name.replace("_"," ").capitalize()}: {val}\n')
    logger.write('\n')
    logger.flush()


def log_dataset_info(datasets, logger):
    # Simply write all dataset details to the logger
    logger.write('Datasets:\n')
    for split in datasets:
        logger.write(f'{split} | ')
        for dataset, metrics in zip(datasets[split]['datasets'], datasets[split]['metrics']):
            logger.write(f'{dataset.dataset_name}')
            logger.write(f' - {dataset.task}')
            for metric in metrics.metrics:
                logger.write(f' - {metric.name}')
            if dataset.num_classes:
                logger.write(f' - {dataset.num_classes} classes')
            logger.write(f' - {dataset.y_size} examples | ')
        logger.write('\n')
    logger.flush()


from collections import Counter
from .metrics import Metric, StringMetric, ElementwiseMetric
import numpy as np
import sklearn.metrics
import torch

class binary_threshold():
    def __init__(self, threshold=0.5):
        self.threshold = threshold
    def __call__(self, y_pred):
        if isinstance(y_pred, torch.Tensor):
            return (y_pred > self.threshold).float()
        elif isinstance(y_pred, np.ndarray):
            return (y_pred > self.threshold).astype(np.float32)
        else:
            return (y_pred > self.threshold)

class Accuracy(ElementwiseMetric):
    def __init__(self, prediction_fn=None, name=None):
        self.prediction_fn = prediction_fn
        if name is None:
            name = 'acc'
        super().__init__(name=name)

    def _compute_element_wise(self, y_pred, y_true):
        if self.prediction_fn is not None:
            y_pred = self.prediction_fn(y_pred)
        return (y_pred==y_true).float()

class F1(Metric):
    def __init__(self, prediction_fn=None, name=None, average='macro', labels=None):
        """
        Calculate F1 score
        Args:
            - prediction_fn: Function to convert y_pred into the same format as y_true (for example, convert logits to max index)
            - name (str): Name of the metric
            - average (str): one of ['binary', 'micro', 'macro', 'weighted', 'samples']
            - labels: The set of labels to include when average != 'binary'  (if None, will use all labels)
        """
        self.prediction_fn = prediction_fn
        self.average = average
        self.labels = labels
        if name is None:
            name = 'F1'
        if average is not None:
            name += f'-{self.average}'
        super().__init__(name=name)

    def _compute(self, y_pred,y_true):
        """
        Args:
            - y_pred: Predicted labels
            - y_true: Ground truth labels
        See https://scikit-learn.org/stable/modules/generated/sklearn.metrics.f1_score.html for further documentation
        """
        if self.prediction_fn is not None:
            y_pred = self.prediction_fn(y_pred)
        score = sklearn.metrics.f1_score(y_true, y_pred, average=self.average, labels=self.labels)
        return torch.tensor(score)

class Precision(Metric):
    def __init__(self, prediction_fn=None, name=None, average='macro', labels=None):
        """
        Calculate Precision
        Args:
            - prediction_fn: Function to convert y_pred into the same format as y_true (for example, convert logits to max index)
            - name (str): Name of the metric
            - average (str): one of ['binary', 'micro', 'macro', 'weighted', 'samples']
            - labels: The set of labels to include when average != 'binary'  (if None, will use all labels)
        """
        self.prediction_fn = prediction_fn
        self.average = average
        self.labels = labels
        if name is None:
            name = 'Precision'
        if average is not None:
            name += f'-{self.average}'
        super().__init__(name=name)

    def _compute(self, y_pred,y_true):
        """
        Args:
            - y_pred: Predicted labels
            - y_true: Ground truth labels
        See https://scikit-learn.org/stable/modules/generated/sklearn.metrics.precision_score.html for further documentation
        """
        if self.prediction_fn is not None:
            y_pred = self.prediction_fn(y_pred)
        score = sklearn.metrics.precision_score(y_true, y_pred, average=self.average, labels=self.labels)
        return torch.tensor(score)

class Recall(Metric):
    def __init__(self, prediction_fn=None, name=None, average='macro', labels=None):
        """
        Calculate Recall
        Args:
            - prediction_fn: Function to convert y_pred into the same format as y_true (for example, convert logits to max index)
            - name (str): Name of the metric
            - average (str): one of ['binary', 'micro', 'macro', 'weighted', 'samples']
            - labels: The set of labels to include when average != 'binary'  (if None, will use all labels)
        """
        self.prediction_fn = prediction_fn
        self.average = average
        self.labels = labels
        if name is None:
            name = 'Recall'
        if average is not None:
            name += f'-{self.average}'
        super().__init__(name=name)

    def _compute(self, y_pred,y_true):
        """
        Args:
            - y_pred: Predicted labels
            - y_true: Ground truth labels
        See https://scikit-learn.org/stable/modules/generated/sklearn.metrics.recall_score.html for further documentation
        """
        if self.prediction_fn is not None:
            y_pred = self.prediction_fn(y_pred)
        score = sklearn.metrics.recall_score(y_true, y_pred, average=self.average, labels=self.labels)
        return torch.tensor(score)

class MultiLabelF1(Metric):
    def __init__(self, prediction_fn=None, name=None, average='micro', labels=None):
        """
        Calculate F1 score for multi-label classification
        Args:
            - prediction_fn: Function to convert y_pred into the same format as y_true (for example, convert logits to max index)
            - name (str): Name of the metric
            - average (str): one of ['binary', 'micro', 'macro', 'weighted', 'samples']
            - labels: The set of labels to include when average != 'binary'  (if None, will use all labels)
        """
        self.prediction_fn = prediction_fn
        self.average = average
        self.labels = labels
        if name is None:
            name = 'F1'
        if average is not None:
            name += f'-{self.average}'
        super().__init__(name=name)

    def _compute(self, y_pred,y_true):
        """
        Args:
            - y_pred: Predicted labels
            - y_true: Ground truth labels
        See https://scikit-learn.org/stable/modules/generated/sklearn.metrics.f1_score.html for further documentation
        """
        if self.prediction_fn is not None:
            y_pred = self.prediction_fn(y_pred)
        score = sklearn.metrics.f1_score(y_true, y_pred, average=self.average, labels=self.labels)
        return torch.tensor(score)

class LRAP(Metric):
    def __init__(self, prediction_fn=None, name=None, labels=None):
        """
        Calculate a ranking-based average precision
        Args:
            - prediction_fn: Function to convert y_pred into the same format as y_true (for example, convert logits to max index)
            - name (str): Name of the metric
            - labels: the set of labels to include (if None, will include all labels)
        """
        self.prediction_fn = prediction_fn
        self.labels = labels
        if name is None:
            name = "LRAP"
        super().__init__(name=name)

    def _compute(self, y_pred, y_true):
        """
        Args:
            - y_pred: Predicted logits
            - y_true: Ground truth
        See https://scikit-learn.org/stable/modules/generated/sklearn.metrics.label_ranking_average_precision_score.html for further documentation
        """
        if self.prediction_fn is not None:
            y_pred = self.prediction_fn(y_pred)
        
        if self.labels:
            # remove samples which do not have a desired label
            filtered_y_pred, filtered_y_true = [], []
            for pred, true in zip(y_pred, y_true):
                if any([true[l] != 0 for l in self.labels]):
                    if isinstance(pred, torch.Tensor):
                        pred = pred.numpy()
                    if isinstance(true, torch.Tensor):
                        true = true.numpy()
                    filtered_y_pred.append(pred)
                    filtered_y_true.append(true)
            y_pred = np.array(filtered_y_pred)
            y_true = np.array(filtered_y_true)
        
        score = sklearn.metrics.label_ranking_average_precision_score(y_true, y_pred)
        return torch.tensor(score)

class MRR(Metric):
    def __init__(self, prediction_fn=None, name=None, labels=None):
        """
        Calculate a variant of the mean reciprocal rank which considers all labels
            If there is only 1 ground truth label, this is equivalent to standard MRR
            For multi-label samples, this still allows for multiple
        Args:
            - prediction_fn: Function to convert y_pred into the same format as y_true (for example, convert logits to max index)
            - name (str): Name of the metric
            - labels: the set of labels to include (if None, will include all labels)
        """
        self.prediction_fn = prediction_fn
        self.labels = labels
        if name is None:
            name = "MRR"
        super().__init__(name=name)

    def _compute(self, y_pred, y_true):
        """
        Args:
            - y_pred: Predicted logits
            - y_true: Ground truth
        """
        
        if self.prediction_fn is not None:
            y_pred = self.prediction_fn(y_pred)
        
        if self.labels:
            # remove samples which do not have a desired label
            filtered_y_pred, filtered_y_true = [], []
            for pred, true in zip(y_pred, y_true):
                if any([true[l] != 0 for l in self.labels]):
                    if isinstance(pred, torch.Tensor):
                        pred = pred.numpy()
                    if isinstance(true, torch.Tensor):
                        true = true.numpy()
                    filtered_y_pred.append(pred)
                    filtered_y_true.append(true)
            y_pred = filtered_y_pred
            y_true = filtered_y_true
        
        reciprocal_ranks = []
        sorted_pred_idxs = np.argsort(-np.array(y_pred), axis=1)
        labels = [np.nonzero(l)[0] for l in y_true]

        for pred_idx, label in zip(sorted_pred_idxs, labels):
            found_labels = 0
            for rank, idx in enumerate(pred_idx):
                if idx in label:
                    reciprocal_ranks.append(1.0 / (rank - found_labels + 1))
                    found_labels += 1

        score = sum(reciprocal_ranks) / len(reciprocal_ranks)
        
        return torch.tensor(score)

class token_F1(StringMetric):
    def __init__(self, prediction_fn=None, name=None, average="macro", unanswerable_phrases=[]):
        """
        Calculate F1 score for token comparisons
        Args:
            - prediction_fn: Function to convert y_pred into the same format as y_true (for example, convert logits to max index)
            - name (str): Name of the metric
        """
        self.prediction_fn = prediction_fn
        self.average = average
        if name is None:
            name = 'token_F1'
        if average is not None:
            name += f'-{self.average}'
        super().__init__(name=name, unanswerable_phrases=unanswerable_phrases)
    
    def _compute(self, y_pred, y_true):
        """
        Args:
            - y_pred (List of str OR List of List of str): Predicted labels
            - y_true (List of str): Ground truth labels
        """
        if self.prediction_fn is not None:
            y_pred = self.prediction_fn(y_pred)

        # Taken from https://github.com/google-research/text-to-text-transfer-transformer/blob/main/t5/evaluation/metrics.py
        def _get_token_f1_macro(y_pred, y_true):
            common_token_counts = (
                Counter(y_true) &
                Counter(y_pred))
            sum_common = sum(common_token_counts.values())
            if len(y_pred) == 0 or len(y_true) == 0:
                return int(y_pred == y_true)
            if sum_common == 0:
                return 0
            precision = 1.0 * sum_common / len(y_pred)
            recall = 1.0 * sum_common / len(y_true)
            f1 = (2 * precision * recall) / (precision + recall)
            return f1

        def _get_token_f1_micro(y_pred, y_true):
            common_token_counts = (
                Counter(y_true) &
                Counter(y_pred))
            sum_common = sum(common_token_counts.values())
            if len(y_pred) == 0 or len(y_true) == 0:
                return int(y_pred == y_true), len(y_pred), len(y_true)
            return sum_common, len(y_pred)-sum_common, len(y_true)-sum_common

        if self.average == "macro":
            f1s = []
            for p, t in zip(y_pred, y_true):

                if isinstance(t, str):
                    f1 = _get_token_f1_macro(p.split(), t.split())
                elif isinstance(t, list):
                    # if multiple ground truths, select the max
                    f1 = self._metric_max_over_ground_truths(_get_token_f1_macro, p.split(), [t_.split() for t_ in t])

                f1s.append(f1)
            return torch.mean(torch.tensor(f1s, dtype=torch.float))
            
        elif self.average == "micro":
            tp, fp, fn = 0, 0, 0
            for pred, true in zip(y_pred, y_true):

                if isinstance(true, str):
                    tp_, fp_, fn_ = _get_token_f1_micro(pred.split(), true.split())
                elif isinstance(true, list):
                    # if multiple ground truths, select the option with highest f1
                    best_tp, best_fp, best_fn, best_f1 = 0, 0, 0, 0
                    for t in true:
                        tp_, fp_, fn_ = _get_token_f1_micro(pred.split(), t.split())
                        f1_ = (2 * tp_) / (2 * tp_ + fp_ + fn_)
                        if f1_ > best_f1:
                            best_tp, best_fp, best_fn = tp_, fp_, fn_
                    tp_ = best_tp
                    fp_ = best_fp
                    fn_ = best_fn

                tp += tp_
                fp += fp_
                fn += fn_
            precision = tp / (tp + fp)
            recall = tp / (tp + fn)
            f1 = 2 * precision * recall / (precision + recall)
            return torch.tensor(f1, dtype=torch.float)
        else:
            raise ValueError(f"Unknown average: {self.average}")

class Exact_Match(StringMetric):
    def __init__(self, prediction_fn=None, name=None, unanswerable_phrases=[], ignore_unanswerable=False):
        """
        Calculate exact match score
        Args:
            - prediction_fn: Function to convert y_pred into the same format as y_true (for example, convert logits to max index)
            - name (str): Name of the metric
        """
        self.prediction_fn = prediction_fn
        if name is None:
            name = 'Exact_Match'
        if ignore_unanswerable:
            name += '_pos_only'
        super().__init__(name=name, unanswerable_phrases=unanswerable_phrases, ignore_unanswerable=ignore_unanswerable)

    def _compute(self, y_pred, y_true):
        """
        Args:
            - y_pred (List of str): Predicted labels
            - y_true (List of str): Ground truth labels
        """
        if self.prediction_fn is not None:
            y_pred = self.prediction_fn(y_pred)

        def _get_exact_match(pred, true):
            return float(pred==true)

        multiple_ground_truths = isinstance(y_true[0],list)

        if multiple_ground_truths:
            matches = [self._metric_max_over_ground_truths(_get_exact_match, p, t) for p, t in zip(y_pred, y_true)]
        else:
            matches = [float(pred == true) for pred, true in zip(y_pred, y_true)]

        return torch.mean(torch.tensor(matches))

class MetricGroup:
    """
    A simple class to group metrics together
    """
    _string_to_class = {
        "f1":F1,
        "precision":Precision,
        "recall":Recall,
        "accuracy":Accuracy,
        "multilabel_f1":MultiLabelF1,
        "label_ranking_average_precision": LRAP,
        "mean_reciprocal_rank": MRR,
        "token_f1":token_F1,
        "exact_match":Exact_Match
    }
    def __init__(self, metrics, **kwargs):
        self.metrics = []
        for metric_str in metrics:
            metric_str = metric_str.lower()
            metric = self._string_to_class[metric_str]
            # allow for multiple variations of a metric
            # eg. F1-micro and F1-macro
            if metric_str in kwargs.keys():
                metric_kwargs = kwargs[metric_str]
                if isinstance(metric_kwargs, list):
                    assert(isinstance(metric_kwargs[0], dict)),"metric kwargs must be dict or list of dicts"
                    for m in metric_kwargs:
                        self.metrics.append(metric(**m))
                elif isinstance(metric_kwargs, dict):
                    self.metrics.append(metric(**metric_kwargs))
            else:
                self.metrics.append(metric())

    def compute(self, y_pred, y_true):
        results = {}
        results_str = ""
        for metric in self.metrics:
            results.update(metric.compute(y_pred, y_true))
            results_str += f'{metric.name}: {results[metric.agg_metric_field]:.4f}\n'
        return results, results_str



from email.mime import base
import os
import csv
import sys

BASE_SOURCE_PATH="{}_{}.{}_seed.{}/{}"
BASE_COTRAINED_PATH="{}_{}_seed.{}/{}"
TASKS={"DailyDialog":[
    # DailyDialog original tasks
    "emotion_recognition", "dialogue_act_classification", "topic_classification",
    # RECCON tasks
    "causal_emotion_span_extraction", "causal_emotion_entailment",
    # CIDER tasks
    "dialogue_nli", "dialogue_reasoning_span_extraction",
    "dialogue_reasoning_multiple_choice_span_selection",
    "dialogue_reasoning_commonsense_relation_prediction",
    # DailyDialog++ task
    "adversarial_response_selection"
    ],
    "Friends": [
        # emoryNLP tasks
        "emotion_recognition", "reading_comprehension", "character_identification",
        "question_answering", "personality_detection",
        # dialogRE task
        "relation_extraction",
        # MELD task
        "MELD_emotion_recognition"
    ]
}

def get_single_result(log_path):
    """
    Get results from single file
    """
    try:
        with open(os.path.join(log_path, "log.txt"), 'r') as f:
            lines = f.readlines()
    except:
        return

    metrics_start = None

    for i, line in enumerate(lines):
        if "Eval on test split" in line:
            metrics_start = i+1
            break
    
    if not metrics_start:
        return
    
    metrics = {}
    for line in lines[metrics_start:]:
        metric_line = line.strip()

        # contine searching until empty line
        if metric_line:
            if len(metric_line.split(": ")) == 2:
                metric, value = metric_line.split(": ")
                metrics[metric] = float(value)
        else:
            break

    return metrics

def get_single_seed_results(training_prefix, model, dataset, seed, fewshot_percent=None):
    """
    Get results for all tasks for a single seed

    Args:
        training_prefix (str): prefix of the training directory, eg. "PRETRAINED_0.1_FEWSHOT" or "COTRAINED"
        model (str): model name
        dataset (str): dataset name
        seed (int): seed number
        fewshot_percent (float): percentage of fewshot data if using fewshot
        tasks (list): list of tasks
    """
    tasks = TASKS[dataset]

    results = {train_type: {target_task: {source_task:{} for source_task in tasks} \
                                for target_task in tasks} \
                    for train_type in ["pretrained_fine-tuned"]}

    # for each task, get the metrics for train/test on source task
    # then get the metrics for train on source, and test on target task
    for source_task in tasks:
        source_prefix = training_prefix if not fewshot_percent else training_prefix + f"_{fewshot_percent}_FEWSHOT"
        source_path = BASE_SOURCE_PATH.format(source_prefix, dataset, source_task, seed, model)
        results['pretrained_fine-tuned'][source_task][source_task] = get_single_result(source_path)
        for target_task in tasks:
            if target_task != source_task:
                target_prefix = "FINETUNED" if not fewshot_percent else f"FINETUNED_{fewshot_percent}_FEWSHOT"
                base_target_path=f"{target_prefix}_{dataset}.{target_task}_seed.{seed}"
                target_path = os.path.join(source_path,base_target_path)
                results['pretrained_fine-tuned'][target_task][source_task] = get_single_result(target_path)

    return results

def get_single_seed_results_cotrained(training_prefix, model, dataset, seed, fewshot_percent=None):
    """
    Get results for all tasks for a single seed

    Args:
        training_prefix (str): prefix of the training directory, eg. "PRETRAINED_0.1_FEWSHOT" or "COTRAINED"
        model (str): model name
        dataset (str): dataset name
        seed (int): seed number
        fewshot_percent (float): percentage of fewshot data if using fewshot
        tasks (list): list of tasks
    """
    tasks = TASKS[dataset]

    results = {train_type: {target_task: {source_task:{} for source_task in tasks} \
                                for target_task in tasks} \
                    for train_type in ["cotrained", "fine-tuned"]}

    # for each task, get the metrics for train/test on source task
    # then get the metrics for train on source, and test on target task
    for source_task in tasks:
        source_prefix = training_prefix if not fewshot_percent else training_prefix + f"_{fewshot_percent}_FEWSHOT"
        source_path = BASE_SOURCE_PATH.format(source_prefix, dataset, source_task, seed, model)
        source_path = source_path.replace("COTRAINED", "PRETRAINED")
        base_result = get_single_result(source_path)
        results['cotrained'][source_task][source_task] = base_result
        results['fine-tuned'][source_task][source_task] = base_result
        # results['cotrained'][source_task][source_task] = {}
        # results['fine-tuned'][source_task][source_task] = {}

        for target_task in tasks:
            if target_task != source_task:
                datasets = f"{dataset}.{source_task}_{dataset}.{target_task}"
                # gather results from cotraining alone
                cotrain_target_prefix = training_prefix if not fewshot_percent else training_prefix + f"_{fewshot_percent}_FEWSHOT"
                cotrain_target_path = BASE_COTRAINED_PATH.format(cotrain_target_prefix, datasets, seed, model)
                results['cotrained'][target_task][source_task] = get_single_result(cotrain_target_path)

                # gather results from fine-tuning after cotraining
                target_prefix = "FINETUNED" if not fewshot_percent else f"FINETUNED_{fewshot_percent}_FEWSHOT"
                base_target_path=f"{target_prefix}_{dataset}.{target_task}_seed.{seed}"
                finetune_target_path = os.path.join(cotrain_target_path,base_target_path)
                results['fine-tuned'][target_task][source_task] = get_single_result(finetune_target_path)

    return results

def print_results(results):
    """
    Print results to terminal
    """
    for k, v in results.items():
        print(k)
        for k1, v1 in v.items():
            print(f"\t{k1}: {v1}")
        print()

def convert_results_to_table(results, aggregation="average"):
    """
    Convert results to table

    Args:
        results (dict): results dictionary
        aggregation (str): aggregation method, either average or sum
    """
    headers = []

    columns = []
    for target_task, source_tasks in results.items():
        headers.append(target_task)
        column = []
        for _, metrics in source_tasks.items():
            if metrics:
                aggregate_value = sum(metrics.values()) if aggregation == "sum" else sum(metrics.values())/len(metrics)
                aggregate_value = round(aggregate_value, 4)
            else:
                aggregate_value = "N/A"
            column.append(aggregate_value)
        columns.append(column)
    return columns,headers

def convert_columns_to_differences(columns):
    for i, column in enumerate(columns):
        for j, value in enumerate(column):
            if i != j:
                if value == "N/A" or columns[i][i] == "N/A":
                    columns[i][j] = "N/A"
                else:
                    columns[i][j] = round(value - columns[i][i],4)
    return columns

def save_results_as_csv(columns, headers, save_path):
    """
    Save results as table in csv format
    """

    rows = list(zip(*columns))
    with open(save_path, 'w') as f:
        writer = csv.writer(f)
        writer.writerow(["Source\Target Task"] + headers)
        for header, row in zip(headers, rows):
            writer.writerow([header, *row])

if __name__ == "__main__":
    if len(sys.argv) < 5:
        print("Usage: python3 results_analysis.py <training_prefix> <model> <dataset> <seed> [<fewshot_percent>]")
        print("Example: python3 results_analysis.py logs_and_models/PRETRAINED t5-base DailyDialog 0.1")
        exit(1)

    training_prefix = sys.argv[1]
    model = sys.argv[2]
    dataset = sys.argv[3]
    seed = sys.argv[4]
    fewshot_percent = None if len(sys.argv) < 6 else float(sys.argv[5])

    if "COTRAINED" in training_prefix:
        results = get_single_seed_results_cotrained(training_prefix, model, dataset, seed, fewshot_percent)
    else:
        results = get_single_seed_results(training_prefix, model, dataset, seed, fewshot_percent)

    for result_type, result in results.items():
        print(f"{result_type} results:")
        print_results(result)

        columns, headers = convert_results_to_table(result)
        base_save_path = f"{training_prefix}_{dataset}_{model}_seed.{seed}_"
        if fewshot_percent:
            base_save_path += f"FEWSHOT.{fewshot_percent}_"
        save_results_as_csv(columns, headers, base_save_path+f"{result_type}_results.csv")
        columns = convert_columns_to_differences(columns)
        save_results_as_csv(columns, headers, base_save_path+f"{result_type}_results_differences.csv")

# Sample usages as Python module

# results = get_single_seed_results("/TLiDB/examples/logs_and_models/PRETRAINED", "t5-base", "DailyDialog", 42)
# columns, headers = convert_results_to_table(results)
# save_results_as_csv(columns, headers,"results.csv")

# columns = convert_columns_to_differences(columns)
# save_results_as_csv(columns, headers,"results_differences.csv")


from .TLiDB_dataset import TLiDB_Dataset, load_split_ids

relations_to_natural_language_map = {
    "per:positive_impression": "positive impression",
    "per:negative_impression": "negative impression",
    "per:acquaintance": "acquaintance",
    "per:alumni": "alumni",
    "per:boss": "boss",
    "per:subordinate": "subordinate",
    "per:client": "client",
    "per:dates": "dates",
    "per:friends": "friends",
    "per:girl/boyfriend": "girlfriend or boyfriend",
    "per:neighbor": "neighbor",
    "per:roommate": "roommate",
    "per:children": "children",
    "per:other_family": "other family",
    "per:parents": "parents",
    "per:siblings": "siblings",
    "per:spouse": "spouse",
    "per:place_of_residence": "place of residence",
    "per:place_of_birth": "place of birth",
    "per:visited_place": "visited place",
    "per:origin": "origin",
    "per:employee_or_member_of": "employee or member of",
    "per:schools_attended": "schools attended",
    "per:works": "works",
    "per:age": "age",
    "per:date_of_birth": "date of birth",
    "per:major": "major",
    "per:place_of_work": "place of work",
    "per:title": "title",
    "per:alternate_names": "alternate names",
    "per:pet": "pet",
    "gpe:residents_of_place": "residents of place",
    "gpe:visitors_of_place": "visitors of place",
    "gpe:births_in_place": "births in place",
    "org:employees_or_members": "employees or members",
    "org:students": "students",
    "unanswerable": "unanswerable",
}

class Friends_dataset(TLiDB_Dataset):
    """
    Friends dataset
    # TODO: Fill in metadata
    
    """

    _dataset_name = "Friends"
    _tasks = [
        'emory_emotion_recognition', 'reading_comprehension', 'character_identification',
        'question_answering', 'personality_detection', 'relation_extraction',
        'MELD_emotion_recognition', 'response_generation'
    ]
    _url = "https://drive.google.com/uc?export=download&id=1QK_XX-d38fKeJlcTcoMT9ku3VZn6bv6H"
    _task_metadatas = {
        "emory_emotion_recognition": {
                "prompt":"emotion:","type":"classification","loader":"utterance_level_classification",
                "collate_type":"classification", "max_decode_tokens":5
        },
        "reading_comprehension": {
            "prompt": "Out of {}, [PLACEHOLDER] is ","type":"span_extraction",
            "loader": "character_span_extraction", "collate_type":"character_span_extraction",
            "max_decode_tokens":20
        },
        "character_identification": {
            "prompt": "Out of {}, '{}' in the phrase '{}' refers to", "type":"classification",
            "loader": "character_identification", "collate_type":"character_identification",
            "max_decode_tokens":5
        },
        "question_answering": {
            "prompt":"","type":"span_extraction","loader":"span_extraction",
            "collate_type":"span_extraction", "max_decode_tokens":64
        },
        "personality_detection": {
            "prompt":"{} is {}:",# <character> is <characteristic>
            "type":"multioutput_classification","num_outputs":5,"num_labels":2,
            "loader":"personality_detection","collate_type":"personality_detection",
            "max_decode_tokens":1, "label_map": {"0":"false", "1":"true"},
            "classes":["aggreeable", "conscientious", "extroverted", "open", "neurotic"]
        },
        "relation_extraction": {
            "prompt":"{} has the following relations with {}:","type":"multilabel_classification",
            "loader":"relation_extraction","collate_type":"relation_extraction",
            "max_decode_tokens":15, "class_to_natural_language_map": relations_to_natural_language_map,
            "default_prediction":"per:positive_impression"
        },
        "MELD_emotion_recognition": {
            "prompt":"emotion:", "type":"classification", "loader":"utterance_level_classification",
            "collate_type":"classification", "max_decode_tokens":5
        },
        "response_generation": {
            "prompt": "", "type":"response_generation","loader":"response_generation",
            "max_decode_tokens":128
        }
    }
    def __init__(self, task, dataset_folder, model_type, split, max_dialogue_length, few_shot_percent=None):
        assert task in self._tasks, f"{task} is not a valid task for {self._dataset_name}"
        super().__init__(self._dataset_name, task, model_type, max_dialogue_length, dataset_folder=dataset_folder)
        self._task_metadata = self._task_metadatas[task]
        self._input_array = []
        self._y_array = []
        self._metadata_fields = []
        self._metadata_array = []
        split_ids = load_split_ids(self._dataset_name, dataset_folder, split, few_shot_percent)
        self._load_data(task, split_ids)
        self._num_classes = len(self.task_labels)
        self._y_size = len(self._y_array)

    def _load_data(self, task, split_ids):
        # get the data loader, based on whether the task is utterance level/dialogue level/span extraction/etc.
        loader = getattr(self, f"_load_{self._task_metadata['loader']}_task")
        return loader(task,split_ids)

    def _load_utterance_level_classification_task(self, task, split_ids):
        for datum in self.dataset['data']:
            if datum['dialogue_id'] in split_ids:
                if task in datum['dialogue_metadata']:
                    dialogue = []
                    for turn in datum['dialogue']:
                        dialogue.append([" ".join(turn['speakers']), turn['utterance']])
                        truncated_dialogue = self._truncate_dialogue(dialogue)
                        str_dialogue = self._convert_dialogue_to_string(truncated_dialogue)
                        if task in turn:
                            self._input_array.append(str_dialogue)
                            self._y_array.append(turn[task])

    def _load_character_span_extraction_task(self, task, split_ids):
        for datum in self.dataset['data']:
            if datum['dialogue_id'] in split_ids:
                if task in datum['dialogue_metadata']:
                    dialogue = [[" ".join(turn['speakers']), turn['utterance']] for turn in datum['dialogue']]
                    truncated_dialogue = self._truncate_dialogue(dialogue)
                    str_dialogue = self._convert_dialogue_to_string(truncated_dialogue)

                    entity_list = datum['dialogue_metadata'][task]['scene_entities']
                    str_entity_list = " or ".join(entity_list)
                    for qa in datum[task]:
                        self._input_array.append({
                            "context": str_dialogue,
                            "question": qa['query'],
                            "entity_list": str_entity_list
                        })
                        answer_start = str_entity_list.find(qa['answer'])
                        assert(answer_start != -1), "couldn't find exact entity name"
                        answer = {
                            "text":qa['answer'],
                            "answer_start":answer_start
                        }
                        self._y_array.append(answer)

    def _load_character_identification_task(self, task, split_ids):
        for datum in self.dataset['data']:
            if datum['dialogue_id'] in split_ids:
                if task in datum['dialogue_metadata']:
                    dialogue = []
                    for turn in datum['dialogue']:
                        dialogue.append([" ".join(turn['speakers']), turn['utterance']])
                        if task in turn:
                            for entity_mention in turn[task]:
                                # add context surrounding the reference
                                post_context_words = turn['utterance'][entity_mention['start']:].split()
                                pre_context_words = turn['utterance'][:entity_mention['start']+len(entity_mention['entity_reference'])+1].split()
                                
                                # prefer whichever context is longer, post context wins in a tie
                                if len(pre_context_words) > len(post_context_words):
                                    context_words = pre_context_words[-4:]
                                else:
                                    context_words = post_context_words[:4]

                                entity_context = " ".join(context_words)

                                self._input_array.append({
                                    "context":dialogue.copy(),
                                    "entity_context": entity_context,
                                    "entity_reference": entity_mention['entity_reference']
                                })
                                self._y_array.append(entity_mention['normalized_entity'])

    def _load_span_extraction_task(self, task, split_ids):

        def _create_truncated_dialogue_for_answer_turn(dialogue, answer_turn):
            """Truncate the dialogue by removing from the far end of the answer span"""
            start_turn_idx = 0
            if self.max_dialogue_length:
                while len(self._convert_dialogue_to_string(dialogue).split()) > self.max_dialogue_length:
                    # if answer is in first half of truncated dialogue, remove a turn from the end
                    if answer_turn-start_turn_idx < len(dialogue)/2:
                        dialogue = dialogue[:-1]
                    # if answer is in second half of truncated dialogue, remove a turn from the beginning
                    else:
                        dialogue = dialogue[1:]
                        start_turn_idx += 1
            return dialogue, start_turn_idx

        for datum in self.dataset['data']:
            if datum['dialogue_id'] in split_ids:
                if task in datum['dialogue_metadata']:
                    dialogue = [[" ".join(turn['speakers']), turn['utterance']] for turn in datum['dialogue']]
                    for qas in datum[task]:
                        for qa in qas['qas']:
                            answers = []
                            for answer in qa['answers']:
                                truncated_dialogue, start_turn_idx = _create_truncated_dialogue_for_answer_turn(dialogue.copy(), answer['answer_utterance_id'])
                                str_dialogue = self._convert_dialogue_to_string(truncated_dialogue)
                                answer_utterance_id = answer['answer_utterance_id']-start_turn_idx
                                
                                prior_turns = self._convert_dialogue_to_string(truncated_dialogue[:answer_utterance_id])
                                
                                # add 1 for space between prior turns and current turn
                                if prior_turns:
                                    answer_start = len(prior_turns) + 1
                                else:
                                    answer_start = 0

                                if not answer['is_speaker']:
                                    if truncated_dialogue[answer_utterance_id][0]:
                                        # if there is a speaker, add speaker + 2 for ': ' between current speaker and current turn
                                        answer_start += len(truncated_dialogue[answer_utterance_id][0]) + 2
                                    answer_start += answer['answer_start']
                                test = str_dialogue[answer_start:answer_start+len(answer['text'])]
                                assert(answer['text'] == test), "answer text doesn't match"
                                answers.append({
                                    "text": answer['text'],
                                    "answer_start": answer_start + len(qa['question']) + 1
                                })
                            self._input_array.append({
                                "context": str_dialogue, "question": qa['question']
                            })
                            self._y_array.append(answers)
                            
    def _load_personality_detection_task(self, task, split_ids):
        for datum in self.dataset['data']:
            if datum['dialogue_id'] in split_ids:
                if task in datum['dialogue_metadata']:
                    for sample in datum[task]:
                        dialogue = [[" ".join(datum['dialogue'][i]['speakers']), datum['dialogue'][i]['utterance']] for i in sample['turns']]
                        truncated_dialogue = self._truncate_dialogue(dialogue)
                        str_dialogue = self._convert_dialogue_to_string(truncated_dialogue)

                        # encoder models can make use of multi-class classification significantly better
                        #   than decoder models, so encoder models get a single sample for all 5 classes
                        if self.model_type == "Encoder":
                            self._input_array.append({
                                "context": str_dialogue,
                                "focus_speaker": sample['focus_speaker']
                            })
                            self._y_array.append(sample['personality_characteristics'])

                        # decoder models get 5 separate samples, 1 for each class
                        else:
                            for characteristic, value in sample['personality_characteristics'].items():
                                self._input_array.append({
                                    "context": str_dialogue,
                                    "focus_speaker": sample['focus_speaker'],
                                    "class": characteristic
                                })
                                self._y_array.append(value)



    def _load_relation_extraction_task(self, task, split_ids):
        for datum in self.dataset['data']:
            if datum['dialogue_id'] in split_ids:
                if task in datum['dialogue_metadata']:
                    for sample in datum[task]:
                        dialogue = [[" ".join(datum['dialogue'][i]['speakers']), datum['dialogue'][i]['utterance']] for i in sample['turns']]
                        truncated_dialogue = self._truncate_dialogue(dialogue)
                        str_dialogue = self._convert_dialogue_to_string(truncated_dialogue)
                        for triple in sample['relation_triples']:
                            self._input_array.append({
                                "context": str_dialogue,
                                "head":triple['head'],
                                "tail":triple['tail']
                            })
                            self._y_array.append(triple['relations'])


    def _collate_encoder(self, batch):
        X, y, metadata = [], [], {}
        for item in batch:
            if self._task_metadata['collate_type'] == "classification":
                X.append(item[0])
            elif self._task_metadata['collate_type'] == "character_span_extraction":
                X.append(self._join_strings(item[0]['entity_list'],"[SEP]",item[0]['question'],"[SEP]",item[0]['context']))
            elif self._task_metadata['collate_type'] == "character_identification":
                context = self._convert_dialogue_to_string(item[0]['context'][::-1])
                prompt = f"Who does '{item[0]['entity_reference']}' refer to in the phrase '{item[0]['entity_context']}'?"
                X.append(self._join_strings(prompt, "[SEP]", context))
            elif self._task_metadata['collate_type'] == "span_extraction":
                X.append(self._join_strings(item[0]['question'],item[0]['context']))
            elif self._task_metadata['collate_type'] == "personality_detection":
                X.append(self._join_strings(item[0]['focus_speaker'],"[SEP]",item[0]['context']))
            elif self._task_metadata['collate_type'] == "relation_extraction":
                X.append(self._join_strings(item[0]['head'],"[SEP]",item[0]['tail'],"[SEP]",item[0]['context']))
            else:
                raise NotImplementedError(f"Collate type {self._task_metadata['collate_type']} not implemented")
            y.append(item[1])
            for k,v in item[2].items():
                if k not in metadata:
                    metadata.append(k)
                metadata[k].append(v)
        return X, y, metadata
        
    def _collate_decoder(self, batch):
        X, y, metadata = [], [], {}
        for item in batch:
            if self._task_metadata['collate_type'] == "classification":
                dialogue = item[0]
                X.append(self._join_strings("context:",dialogue,self._task_metadata['prompt']))
            elif self._task_metadata['collate_type'] == "character_span_extraction":
                X.append(self._join_strings("context:",item[0]['context'],"question:",item[0]['question'],self._task_metadata['prompt'].format(item[0]['entity_list'])))
            elif self._task_metadata['collate_type'] == "character_identification":
                truncated_context = self._truncate_dialogue(item[0]['context'])
                context = self._convert_dialogue_to_string(truncated_context)
                prompt = self._task_metadata['prompt'].format(" or ".join(self.task_labels), item[0]['entity_reference'],item[0]['entity_context'])
                X.append(self._join_strings("context:",context,prompt))
            elif self._task_metadata['collate_type'] == "span_extraction":
                X.append(self._join_strings("context:",item[0]['context'],
                                            "question:", item[0]['question'], "answer:"))
            elif self._task_metadata['collate_type'] == "personality_detection":
                prompt = self._task_metadata['prompt'].format(item[0]['focus_speaker'], item[0]['class'])
                X.append(self._join_strings("context:",item[0]['context'],prompt))
            elif self._task_metadata['collate_type'] == "relation_extraction":
                prompt = self._task_metadata['prompt'].format(item[0]['head'], item[0]['tail'])
                X.append(self._join_strings("context:",item[0]['context'],prompt))
            else:
                raise NotImplementedError(f"Collate type {self._task_metadata['collate_type']} not implemented")
            y.append(item[1])
            for k,v in item[2].items():
                if k not in metadata:
                    metadata.append(k)
                metadata[k].append(v)
        labels = self.task_labels
        if labels:
            metadata['labels'] = labels
        return X, y, metadata

    def _collate_encoderdecoder(self, batch):
        X, y, metadata = [], [], {}
        for item in batch:
            if self._task_metadata['collate_type'] == "classification":
                dialogue = item[0]
                X.append(self._join_strings("context:",dialogue,self._task_metadata['prompt']))
            elif self._task_metadata['collate_type'] == "character_span_extraction":
                X.append(self._join_strings("context:",item[0]['context'],"question:",item[0]['question'],self._task_metadata['prompt'].format(item[0]['entity_list'])))
            elif self._task_metadata['collate_type'] == "character_identification":
                truncated_context = self._truncate_dialogue(item[0]['context'])
                context = self._convert_dialogue_to_string(truncated_context)
                prompt = self._task_metadata['prompt'].format(" or ".join(self.task_labels), item[0]['entity_reference'],item[0]['entity_context'])
                X.append(self._join_strings("context:",context,prompt))
            elif self._task_metadata['collate_type'] == "span_extraction":
                X.append(self._join_strings("context:",item[0]['context'],
                                            "question:", item[0]['question'], "answer:"))
            elif self._task_metadata['collate_type'] == "personality_detection":
                prompt = self._task_metadata['prompt'].format(item[0]['focus_speaker'], item[0]['class'])
                X.append(self._join_strings("context:",item[0]['context'],prompt))
            elif self._task_metadata['collate_type'] == "relation_extraction":
                prompt = self._task_metadata['prompt'].format(item[0]['head'], item[0]['tail'])
                X.append(self._join_strings("context:",item[0]['context'],prompt))
            else:
                raise NotImplementedError(f"Collate type {self._task_metadata['collate_type']} not implemented")
            y.append(item[1])
            for k,v in item[2].items():
                if k not in metadata:
                    metadata.append(k)
                metadata[k].append(v)
        labels = self.task_labels
        if labels:
            metadata['labels'] = labels
        return X, y, metadata


import torch
from itertools import chain

class TLiDB_model:
    def __init__(self, config):
        self.config = config

    @property
    def model(self):
        """
        Base model underlying the encoding/decoding
        model is of type transformers.PreTrainedModel
        """
        return self._model
    
    @model.setter
    def model(self, model):
        self._model = model

    @property
    def layers(self):
        return self._layers

    @layers.setter
    def layers(self, layers):
        self._layers = layers

    @layers.getter
    def layers(self):
        return self._layers

    def parameters(self):
        """Convenience function to gather all model parameters
            to be passed to a torch.optim optimizer"""
        params = []
        for layer_name, layer in self.layers.items():
            params.extend(layer.parameters())

        return params

    def named_parameters(self):
        """Convenience function to gather all named model parameters
            to be passed to a torch.optim optimizer"""
        named_params = chain()
        for layer_name, layer in self.layers.items():
            named_params = chain(named_params, layer.named_parameters())
        return named_params

    def state_dict(self):
        state_dict = {}
        for layer_name, layer in self.layers.items():
            state_dict[layer_name] = layer.state_dict()
        return state_dict

    def load_state_dict(self, state_dict):
        """Convenience function to load state dict for all layers"""
        return NotImplementedError

    def zero_grad(self, set_to_none=True):
        """Convenience function to zero gradients for all layers"""
        for layer_name, layer in self.layers.items():
            layer.zero_grad(set_to_none=set_to_none)

    def train(self, mode=True):
        """Convenience function to set all layers to train mode"""
        for layer_name, layer in self.layers.items():
            layer.train(mode)

    def init_weights(self):
        for layer_name, layer in self.layers.items():
            # main model is always Transformers-based and does it's own init
            if layer_name == "model":
                pass
            elif isinstance(layer, torch.nn.Module):
                torch.nn.init.xavier_uniform_(layer.weight)
            else:
                raise ValueError("Layer {} is not a torch.nn.Module or Transformers Model".format(layer_name))

    @property
    def forward(self):
        return NotImplementedError

    @forward.setter
    def forward(self, forward):
        self._forward = forward

    def __call__(self, *args, **kwargs):
        return self._forward(*args, **kwargs)

    def transform_inputs(self, inputs):
        return NotImplementedError

    def transform_outputs(self, outputs):
        return NotImplementedError

    def to(self, device):
        """Convenience function to move all layers to a device"""
        for layer_name, layer in self.layers.items():
            layer.to(device)



from .TLiDB_dataset import TLiDB_Dataset
from TLiDB.metrics.all_metrics import Accuracy

class clinc150_dataset(TLiDB_Dataset):
    """
    CLINC150 dataset
    This is the full dataset from https://github.com/clinc/oos-eval

    Input (x):
        - text (str): Text utterance

    Target (y):
        - label (list): List of [Domain, Intent] labels

    Metadata:
        - domain (str): Domain of the utterance

    """
    _dataset_name = "clinc150"
    _tasks = ["intent_detection"]
    _url = "https://drive.google.com/uc?export=download&id=1dG6KXQ6L7xpbnWmhW9Xo3vPSfYstk43E"
    def __init__(self, task, dataset_folder, model_type, split=None):
        assert task in self._tasks, f"{task} is not a valid task for {self._dataset_name}"
        super().__init__(self.dataset_name, task, model_type, dataset_folder=dataset_folder)
        
        # initialize task data and metadata
        categories = [
            "auto and commute","banking","credit cards","home",
            "kitchen and dining","meta","small talk","travel",
            "utility","work"
            ]
        self._input_array = []
        self._y_array = []
        self._metadata_fields = ["domains"]
        self._metadata_array = [[] for _ in self._metadata_fields]
        
        # convert labels to human readable
        labels = [label.replace("_"," ") for label in self.task_labels]
        formatted_labels = []
        for label in labels:
            for c in categories:
                if c == label[:len(c)]:
                    formatted_label = c+":"+label[len(c):]
                    formatted_labels.append(formatted_label)
        self.task_labels = formatted_labels

        for datum in self.dataset['data']:
            if split and datum['dialogue_metadata']['original_data_partition'] != split:
                continue
            utterance = datum['dialogue'][0]
            domain = utterance['intent_detection']['domain']
            intent = utterance['intent_detection']['intent']
            self._input_array.append(utterance['utterance'])
            self._y_array.append([domain, intent])
            self.get_metadata_field("domains").append(domain)

        self._num_classes = len(self.task_labels)
        self._y_size = len(self._y_array)

    def get_input(self, idx):
        return self._input_array[idx]

    def get_metadata(self, idx):
        return {
            "domains": self.get_metadata_field("domains")[idx],
            }

    def _collate_encoder(self, batch):
        X,y, metadata = [], [], {}
        for item in batch:
            X.append(item[0])
            y.append(f"{item[1][0].replace('_',' ')}: {item[1][1].replace('_',' ')}")
            for k, v in item[2].items():
                if k not in metadata:
                    metadata[k] = []
                metadata[k].append(v)
        return X,y, metadata

    def _collate_decoder(self, batch):
        X,y, metadata = [], [], {}
        for item in batch:
            X.append(item[0])
            y.append(f"{item[1][0].replace('_',' ')}: {item[1][1].replace('_',' ')}")
            for k, v in item[2].items():
                if k not in metadata:
                    metadata[k] = []
                metadata[k].append(v)
        labels = self.task_labels
        if labels:
            metadata['labels'] = labels
        return X,y, metadata

    def _collate_encoderdecoder(self, batch):
        X,y, metadata = [], [], {}
        for item in batch:
            X.append(item[0])
            y.append(f"{item[1][0].replace('_',' ')}: {item[1][1].replace('_',' ')}")
            for k, v in item[2].items():
                if k not in metadata:
                    metadata[k] = []
                metadata[k].append(v)
        labels = self.task_labels
        if labels:
            metadata['labels'] = labels
        return X,y, metadata


import sys
import os
package_directory = os.path.dirname(os.path.abspath(__file__))
TLiDB_FOLDER = os.path.join(package_directory, "..")
sys.path.append(TLiDB_FOLDER)
import torch.nn as nn
from TLiDB.metrics.loss import ElementwiseLoss


class LM_CrossEntropyLoss:
    def __init__(self, reduction="none", ignore_index=-100):
        self.reduction = reduction
        self.ignore_index = ignore_index
    def __call__(self, output, target):
        output = output.view(-1, output.size(-1))
        target = target.view(-1)
        return nn.CrossEntropyLoss(reduction=self.reduction, ignore_index=self.ignore_index)(output, target)

def initialize_loss(loss_function):
    if loss_function == "cross_entropy":
        return ElementwiseLoss(loss_fn = nn.CrossEntropyLoss(reduction="none"))
    elif loss_function == "LM_cross_entropy":
        return ElementwiseLoss(loss_fn = LM_CrossEntropyLoss(ignore_index=-100))
    elif loss_function == "BCE_with_logits":
        return ElementwiseLoss(loss_fn = nn.BCEWithLogitsLoss(reduction="none"))
    else:
        raise ValueError(f"Unknown loss function: {loss_function}")



import torch

def numel(obj):
    if torch.is_tensor(obj):
        return obj.numel()
    elif isinstance(obj, list):
        return len(obj)
    else:
        raise TypeError("Invalid type for numel")


import TLiDB
# dataset url can be found in the google drive, where the original link is:
#   https://drive.google.com/file/d/1sqaiYTm9b9SPEzehdjp_DXEovVId6Fvq/view?usp=sharing
# and needs to be reformatted as:
#   https://drive.google.com/uc?export=download&id=1sqaiYTm9b9SPEzehdjp_DXEovVId6Fvq

def get_dataset(dataset, task, dataset_folder, **dataset_kwargs):
    """
    Returns the appropriate TLiDB dataset class
    Args:
        dataset: name of the dataset
        dataset_kwargs: keyword arguments to pass to the dataset class
    Returns:
        dataset: TLiDB dataset class
    """
    if dataset not in TLiDB.supported_datasets:
        raise ValueError(f"{dataset} is not a supported dataset, must be one of: {TLiDB.supported_datasets}")

    if dataset == "DailyDialog":
        from TLiDB.datasets.DailyDialog_dataset import DailyDialog_dataset as dataset_class
    elif dataset == "Friends":
        from TLiDB.datasets.Friends_dataset import Friends_dataset as dataset_class
    elif dataset == "clinc150":
        from TLiDB.datasets.clinc150_dataset import clinc150_dataset as dataset_class
    else:
        raise ValueError(f"{dataset} is not a supported dataset, must be one of: {TLiDB.supported_datasets}")
        
    return dataset_class(task, dataset_folder, **dataset_kwargs)


from tqdm import tqdm
import torch
from examples.utils import detach_and_clone, collate_list, concat_t_d, save_algorithm_if_needed, save_pred_if_needed
from TLiDB.data_loaders.data_loaders import TLiDB_DataLoader

def run_epoch(algorithm, datasets, config, logger, train):
    """
    Run one epoch of training or validation.
    Args:
        algorithm: (Algorithm) the algorithm to run
        datasets: (dict) contains all information about the datasets: splits, losses, etc.
        config: (Config) the configuration
        logger: (Logger) the logger
        train: (boolean) True for training, False for validation (in val mode).
    """
    if train:
        algorithm.train()
        torch.set_grad_enabled(True)
    else:
        algorithm.eval()
        torch.set_grad_enabled(False)
    
    # convert all datasets into a single multi-task and multi-domain dataloader
    dataloader = TLiDB_DataLoader(datasets)
    task_datasets = [concat_t_d(d.task, d.dataset_name) for d in datasets['datasets']]

    epoch_y_true = {t_d: [] for t_d in task_datasets}
    epoch_y_pred = {t_d: [] for t_d in task_datasets}

    # Using enumerate(iterator) can sometimes leak memory in some environments (!)
    # so we manually increment the step
    pbar = tqdm(dataloader) if config.progress_bar else dataloader
    # cumulative loss during the epoch
    total_loss = {t_d: 0 for t_d in task_datasets}
    step = {t_d: 0 for t_d in task_datasets}
    for batch in pbar:
        _, _, batch_metadata = batch
        batch_t_d = concat_t_d(batch_metadata['task'],batch_metadata['dataset_name'])
        if train:
            batch_results = algorithm.update(batch,sum(step.values()))
        else:
            batch_results = algorithm.evaluate(batch)
        
        # These should already be detached, but in some versions they won't get garbage
        #   collected properly if not detached again
        epoch_y_true[batch_t_d].append(detach_and_clone(batch_results['y_true']))
        y_pred = detach_and_clone(batch_results['y_pred'])
        
        epoch_y_pred[batch_t_d].append(y_pred)

        total_loss[batch_t_d] += detach_and_clone(batch_results['objective']['loss_value'])
        desc = "Train losses" if train else "Validation losses"
        for t_d in task_datasets:
            desc += f" | {t_d}: {total_loss[t_d]/(step[t_d]+1):0.4f}"

        pbar.set_description(desc)
        step[batch_t_d] += 1

        # TODO: Option for 'log every n steps'

    for t_d in task_datasets:
        epoch_y_true[t_d] = collate_list(epoch_y_true[t_d])
        epoch_y_pred[t_d] = collate_list(epoch_y_pred[t_d])


    # This loop is determined by the model/task/mode(train/val)
    results = {}
    if algorithm.requires_metric_calculation():
        logger.write('Epoch eval:\n')
        for m, d in zip(datasets['metrics'],datasets['datasets']):
            t_d = concat_t_d(d.task,d.dataset_name)
            r, r_str = m.compute(epoch_y_pred[t_d], epoch_y_true[t_d])
            results[t_d] = r
            logger.write(f"{d.dataset_name} {d.task}-\n{r_str}\n")

    return results, epoch_y_pred


def train(algorithm, datasets, config, logger, epoch_offset, best_val_metric):
    for epoch in range(epoch_offset, config.num_epochs):
        logger.write(f'\nEpoch {epoch}\n')
        # train
        run_epoch(algorithm, datasets['train'], config, logger, train=True)

        # allow for training without dev set, will not save model
        if not datasets['dev'].get('datasets', None):
            continue

        # evaluate on validation set
        val_results, y_pred = run_epoch(algorithm, datasets['dev'], config, logger, train=False)
        val_metrics = [val_results[d][m] for d in val_results for m in val_results[d]]
        cur_val_metric = sum(val_metrics)/len(val_metrics)
        logger.write(f'Validation metric: {cur_val_metric:0.4f}\n')

        if best_val_metric is None:
            is_best=True
        else:
            is_best = cur_val_metric > best_val_metric

        if is_best:
            best_val_metric = cur_val_metric
            logger.write(f'Epoch {epoch} gives best validation result so far.\n')

        # save algorithm and model
        save_algorithm_if_needed(algorithm, epoch, config, best_val_metric, is_best, logger)
        # save predictions
        save_pred_if_needed(y_pred, epoch, config, is_best, config.save_path_dir)

        logger.write('\n')
        logger.flush()


def evaluate(algorithm, datasets, config, logger, epoch, is_best):
    algorithm.eval()
    torch.set_grad_enabled(False)
    for split in datasets:
        for dataset, loader, metric in zip(datasets[split]['datasets'], datasets[split]['loaders'], datasets[split]['metrics']):
            epoch_y_true = []
            epoch_y_pred = []

            pbar = tqdm(iter(loader)) if config.progress_bar else iter(loader)

            for batch in pbar:
                # add batch metadata to the batch
                X, y, batch_metadata = batch
                batch_metadata['task'] = dataset.task
                batch_metadata['dataset_name'] = dataset.dataset_name
                batch_metadata['task_metadata'] = dataset.task_metadata
                batch = (X, y, batch_metadata)

                batch_results = algorithm.evaluate(batch)
                epoch_y_true.append(detach_and_clone(batch_results['y_true']))
                y_pred = detach_and_clone(batch_results['y_pred'])
                epoch_y_pred.append(y_pred)

            epoch_y_pred = collate_list(epoch_y_pred)
            epoch_y_true = collate_list(epoch_y_true)

            r, r_str = metric.compute(epoch_y_pred, epoch_y_true)
            r['epoch'] = epoch
            logger.write(f"Eval on {split} split at epoch {epoch}: {dataset.dataset_name} {dataset.task}-\n{r_str}\n")

            # skip saving train data as the dataloader will shuffle data
            if split != "train":
                save_pred_if_needed(y_pred, epoch, config, is_best, config.save_path_dir)



import setuptools
import os
import sys

here = os.path.abspath(os.path.dirname(__file__))
sys.path.insert(0, os.path.join(here, "TLiDB"))


with open("README.md", "r", encoding="utf-8") as f:
    long_description = f.read()

setuptools.setup(
    name="tlidb",
    version="0.0.2",
    license='MIT',
    url='https://github.com/alon-albalak/TLiDB',
    author="Alon Albalak",
    author_email="alon_albalak@ucsb.edu",
    description="The Transfer Learning in Dialogue Baselines Toolkit",
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires = [
        'nltk==3.6.5',
        'scikit-learn==1.0',
        'transformers==4.11.3',
        'torch==1.10',
        'sentencepiece==0.1.96',
    ],
    packages=setuptools.find_packages(exclude=['dataset_preprocessing']),
    python_requires='>=3.6',
    classifiers=[
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Intended Audience :: Science/Research',
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
    ],
)


from TLiDB.metrics.metrics import Metric, ElementwiseMetric

class Loss(Metric):
    def __init__(self, loss_fn, name=None):
        self.loss_fn = loss_fn
        if name is None:
            name = 'loss'
        super().__init__(name=name)

    def _compute(self, y_pred, y_true):
        """
        Helper for computing element-wise metric, implemented for each metric
        Args:
            - y_pred (Tensor): Predicted targets or model output
            - y_true (Tensor): True targets
        Output:
            - element_wise_metrics (Tensor): tensor of size (batch_size, )
        """
        return self.loss_fn(y_pred, y_true)

class ElementwiseLoss(ElementwiseMetric):
    def __init__(self, loss_fn, name=None):
        self.loss_fn = loss_fn
        if name is None:
            name = 'loss'
        super().__init__(name=name)

    def _compute_element_wise(self, y_pred, y_true):
        """
        Helper for computing element-wise metric, implemented for each metric
        Args:
            - y_pred (Tensor): Predicted targets or model output
            - y_true (Tensor): True targets
        Output:
            - element_wise_metrics (Tensor): tensor of size (batch_size, )
        """
        return self.loss_fn(y_pred, y_true)



from utils import move_to
from losses import initialize_loss
from .algorithm import Algorithm
import torch

def multiclass_logits_to_pred(logits):
    """
    Takes multi-class logits of size (batch_size, ..., n_classes) and returns predictions
    by taking an argmax at the last dimension
    """
    assert logits.dim() > 1
    return logits.argmax(-1)

class EncoderAlgorithm(Algorithm):
    def __init__(self, config, datasets):
        super().__init__(config, datasets)

    def process_batch(self, batch):
        """
        A helper function for update() and evaluate() that process the batch
        Args:
            - batch: a batch of data yielded by the DataLoader
        Output:
            - results: a dictionary of results
                - y_pred: the prediction of the model
                - y_true: the ground truth
                - metadata: the metadata of the batch
        """
        X, y_true, metadata = batch
        task_type=metadata['task_metadata']['type']

        # task-specific preprocessing
        X, y_true, metadata = getattr(self, f"_{task_type}_preprocessing")(X, y_true, metadata)

        X = self.model.transform_inputs(X, metadata)
        transformed_y_true = self.model.transform_outputs(X, y_true, task_type, metadata)

        X = move_to(X, self.device)
        transformed_y_true = move_to(transformed_y_true, self.device)

        outputs = self.model(X,metadata['task'],metadata['dataset_name'])

        # task-specific loss calculation
        loss = getattr(self, f"_calculate_{task_type}_loss")(outputs, transformed_y_true, metadata, return_dict=False)

        # task-specific postprocessing
        y_pred, y_true, metadata = getattr(self, f"_{task_type}_postprocessing")(X, outputs, y_true, transformed_y_true, metadata)

        results = {
            'y_pred': y_pred,
            'y_true': y_true,
            'metadata': metadata,
            "objective": {"loss_name": "cross_entropy"}
        }

        results['objective']['loss_value'] = loss.item()

        return results, loss

    def requires_metric_calculation(self):
        return True

    def _classification_preprocessing(self, X, y_true, metadata):
        X = [self.replace_sep_token(x) for x in X]
        return X, y_true, metadata

    def _classification_postprocessing(self, X, outputs, y_true, transformed_y_true, metadata):
        y_pred = multiclass_logits_to_pred(outputs)
        return y_pred, transformed_y_true, metadata

    def _calculate_classification_loss(self, outputs, y_true, metadata, return_dict=False):
        metric = initialize_loss("cross_entropy")
        loss = metric.compute(outputs, y_true, return_dict=return_dict)
        return loss

    def _multioutput_classification_preprocessing(self, X, y_true, metadata):
        X = [self.replace_sep_token(x) for x in X]
        return X, y_true, metadata

    def _multioutput_classification_postprocessing(self, X, outputs, y_true, transformed_y_true, metadata):
        # first, flatten outputs, then calculate predictions, then reshape
        outputs = outputs.reshape([-1, metadata['task_metadata']['num_labels']])
        y_pred = multiclass_logits_to_pred(outputs)
        y_pred = y_pred.reshape(transformed_y_true.shape)
        return y_pred, transformed_y_true, metadata

    def _calculate_multioutput_classification_loss(self, outputs, y_true, metadata, return_dict=False):
        # flatten the outputs and targets
        metric = initialize_loss("cross_entropy")
        outputs = outputs.reshape([-1, metadata['task_metadata']['num_labels']])
        y_true = y_true.flatten()
        loss = metric.compute(outputs, y_true, return_dict=return_dict)
        return loss

    def _multilabel_classification_preprocessing(self, X, y_true, metadata):
        X = [self.replace_sep_token(x) for x in X]
        return X, y_true, metadata

    def _multilabel_classification_postprocessing(self, X, outputs, y_true, transformed_y_true, metadata):
        """
        Transforms the logits output into a multi-label prediction

        First, we take the logits and apply a softmax to get the probabilities
        Then, we take the probabilities and apply a threshold to get the binary decision
        For each sample, if no class is over the threshold, we predict the class with the highest probability

        This method works well for multilabel problems with 1 or 2 correct labels at a time, likely will need to be tuned for larger quantities
        """

        y_pred = torch.softmax(outputs, dim=-1)
        binarized_y_pred = (y_pred > 0.1).float()
        top_indices = torch.argmax(y_pred, dim=-1)
        preds = []
        for y, binary_y, ind in zip(y_pred, binarized_y_pred, top_indices):
          if torch.sum(binary_y) == 0:
            pred = [0 for _ in range(len(y))]
            pred[ind] = 1
            pred = torch.tensor(pred, dtype=torch.float, device=y_pred.device)
            preds.append(pred)
          else:
            preds.append(binary_y)
        y_pred = torch.stack(preds)

        return y_pred, transformed_y_true, metadata

    def _calculate_multilabel_classification_loss(self, outputs, y_true, metadata, return_dict=False):
        metric = initialize_loss("BCE_with_logits")
        loss = metric.compute(outputs, y_true, return_dict=return_dict)
        return loss

    def _span_extraction_preprocessing(self, X, y_true, metadata):
        # make the tokenizer return a token offset mapping
        metadata['return_offsets_mapping'] = True
        return X, y_true, metadata

    def _span_extraction_postprocessing(self, X, outputs, y_true, transformed_y_true, metadata):
        del metadata['return_offsets_mapping']
        assert outputs.dim() == 3

        # decode the predictions
        y_pred_tokens = []
        pred_positions = outputs.argmax(1).tolist()
        for input_ids, (start_pred, end_pred) in zip(X.input_ids, pred_positions):
            y_pred_tokens.append(input_ids[start_pred:end_pred+1])
        y_pred = self.model.tokenizer.batch_decode(y_pred_tokens, skip_special_tokens=True)
        
        # tokenize the ground truth
        if isinstance(y_true[0], list):
            tokenized_y_true = []
            for answers in y_true:
                tokenized_answers = self.model.tokenizer([a['text'] for a in answers])
                tokenized_y_true.append(self.model.tokenizer.batch_decode(tokenized_answers.input_ids, skip_special_tokens=True))

        else:
            tokenized_answers = self.model.tokenizer([a['text'] for a in y_true])
            tokenized_y_true = self.model.tokenizer.batch_decode(tokenized_answers.input_ids, skip_special_tokens=True)

        return y_pred, tokenized_y_true, metadata

    def _calculate_span_extraction_loss(self, outputs, y_true, metadata, return_dict=False):
        start_logits, end_logits = outputs.split(1, dim=-1)
        start_logits = start_logits.squeeze(-1).contiguous()
        end_logits = end_logits.squeeze(-1).contiguous()

        ignored_index = start_logits.size(1)
        start_positions,end_positions=y_true
        start_positions = start_positions.clamp(0, ignored_index)
        end_positions = end_positions.clamp(0, ignored_index)

        metric = initialize_loss("cross_entropy")
        start_loss = metric.compute(start_logits, start_positions, return_dict=return_dict)
        end_loss = metric.compute(end_logits, end_positions, return_dict=return_dict)
        loss = (start_loss + end_loss)/2

        return loss

    def _multiple_choice_preprocessing(self, X, y_true, metadata):
        # needs to unsqueeze the inputs
        X = [self.replace_sep_token(x) for q in X for x in q]

        # keep outputs as indices
        y_true = [int(y) for y in y_true]
        return X, y_true, metadata

    def _multiple_choice_postprocessing(self, X, outputs, y_true, transformed_y_true, metadata):
        # needs to group the data back together to make a single prediction
        outputs = outputs.view(-1, metadata['task_metadata']['num_choices'])
        y_pred = multiclass_logits_to_pred(outputs)
        return y_pred, transformed_y_true, metadata

    def _calculate_multiple_choice_loss(self, outputs, y_true, metadata, return_dict=False):
        outputs = outputs.view(-1, metadata['task_metadata']['num_choices'])
        metric = initialize_loss("cross_entropy")
        loss = metric.compute(outputs, y_true, return_dict=return_dict)
        return loss

    def replace_sep_token(self, string):
        return string.replace("[SEP]",self.model.tokenizer.sep_token)
